(*---------------------------------------------------------------------------

   Generating comPositional UNITY from generalUNITY.

   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.uu.nl

   Last update 
        March 2002, Taupo 6
	Feb 2002, Taupo 6
	Apr 99
---------------------------------------------------------------------------*)           


(*---------------------------------------------------------------------------

  USER SECTION 0: Naming your theory.
  
---------------------------------------------------------------------------*)           

val give_a_name_of_your_concrete_UNITY = "com"







(*---------------------------------------------------------------------------

   PRELUDE: 

   This section loads the minimum libraries you need. If you need more
   in order to prove the primitive properties of your concrete UNITY,
   then you can load them here.

---------------------------------------------------------------------------*)           

(* Uncomment this if you want to compile this file. *)

open Parse HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;



val _ = new_theory (give_a_name_of_your_concrete_UNITY ^ "UNITY") ;


(* Comment all these load commands if you want to compile this file. 

load "bossLib" ;
load "predTheory" ;
load "predLib" ;
load "combinTheory" ;
load "actionTheory" ;
load "generalUNITYTheory" ;
load "uprogTheory" ;

*)

open bossLib ;
open predTheory ;
open predLib ;
open PSyntax ;
open actionTheory ;
open generalUNITYTheory ;
open uprogTheory ;

val _ = associate_restriction ("!!", "RES_qAND") ;
val _ = associate_restriction ("??", "RES_qOR") ;


(*---------------------------------------------------------------------------

USER SECTION 1: Defining the concrete UNITY operators.

   In this section you'll have to define your concrete UNITY
   operators, namely:

	*implies
	*unless
	*ensures
	*leadsto
 
   There are also a number of ML variables which you have to set
   because the wizard need the information from you. This are, for
   example, variables containing information about how you call your
   unless operator and under which name its definition is stored.

----------------------------------------------------------------------------*)

(*---------------------------------------------------------------------------

Some ML names for type abbreviations. You don't have to have
them... but they're often handy.

----------------------------------------------------------------------------*)

(* type of state predicates *)
val Pred    = ty_antiq (Type `:('a->bool)`) ;

(* type of general UNITY operators *)
val UOP     = ty_antiq (Type `:('a->bool)->('a->bool)->bool`) ;

(* type of actions *)
val Action  = ty_antiq (Type `:'a->'a->bool`) ;

(* type of sets of actions *)

val Actions = ty_antiq (Type `:^(type_of Action)->bool`) ;

(* type of UNITY programs *)

val Prog    = Actions

(* type of comUNITY operators *)

val comUOP    = ty_antiq 
		(Type `:^(type_of Prog)->
			^(type_of Pred)->
			(^(type_of Pred) -> bool) ->
			^(type_of UOP)`) ;


(*---------------------------------------------------------------------------

Defining implies

----------------------------------------------------------------------------*)


val IMPLIES_def = Define 
	`IMPLIES J A p q 
	 =
	 A p /\ A q /\ 
	 valid (J AND p IMP q)` ;

(*---------------------------------------------------------------------------

Defining unless

----------------------------------------------------------------------------*)

val UNLESS_def = Define
	`UNLESS P J A p q
	 =
	 STABLE P J /\ A p /\ A q /\
	 (!a. P a ==> HOA (J AND p AND NOT q) a (p OR q))` ;

(*---------------------------------------------------------------------------

Defining ensures

----------------------------------------------------------------------------*)

val ENSURES_def = Define
	`ENSURES P J A p q
	 =
	 UNLESS P J A p q /\
	 (?a. P a ==> HOA (J AND p AND NOT q) a q)` ;


(*---------------------------------------------------------------------------

  Modify the information assigned to the following variables:

----------------------------------------------------------------------------*)

val implies_name   = "IMPLIES" ;
val unless_name    = "UNLESS" ;
val ensures_name   = "ENSURES" ;

val implies_def    = IMPLIES_def ;
val unless_def     = UNLESS_def ;
val ensures_def    = ENSURES_def ;

val implies_object = --`(IMPLIES J A):^UOP`-- ;
val unless_object  = --`(UNLESS:^comUOP) P J A`-- ;
val ensures_object = --`(ENSURES:^comUOP) P J A`-- ;

val implies_other_required_defs = [] ;
val unless_other_required_defs  = [HOA_def, STABLE_def] ;
val ensures_other_required_defs = [HOA_def, STABLE_def] ;


(*---------------------------------------------------------------------------

Defining leadsto. It should be defined as the GLEADSTO of your
concrete ensures.

>From the definition of leadsto below, change only the name of the
leadsto and the name of its _def. So for example you may want to call
it FOO instead of LEADSTO, and have the definition saved under the
name FOO_def instead of LEADSTO_def.

----------------------------------------------------------------------------*)

val LEADSTO_def = Define `LEADSTO P J A = GLEADSTO ^ensures_object` ;

(*---------------------------------------------------------------------------

  Modify the information assigned to the following variables:

----------------------------------------------------------------------------*)

val leadsto_name   = "LEADSTO" ;
val leadsto_def    = LEADSTO_def ;
val leadsto_object = --`(LEADSTO:^comUOP) P J A`-- ;
val leadsto_other_required_defs = [ENSURES_def] ;


(*---------------------------------------------------------------------------

USER SECTION 2: Specifying which laws to generate.

In the lists below, you should specify which laws exported by
generalUNITYTheory that you want to have in your concrete UNITY.

For each law you want to have, you should specify:

	(new_name_of_the_law, old_name_of_the_law)

----------------------------------------------------------------------------*)

val unless_laws_to_be_generated =

	[("_includes_implies_thm", implies_to_unless_thm),
	 ("_includes_IMP_thm", IMP_to_unless_thm),
	 ("_REFL_thm", unless_REFL_thm),
	 ("_post_weakening_thm", unless_post_weakening_thm),
	 ("_simple_conjunction_thm", unless_simple_conjunction_thm),
	 ("_simple_disjunction_thm", unless_simple_disjunction_thm)
	] ;

val ensures_laws_to_be_generated : ((string * thm) list) = [] ;

val leadsto_laws_to_be_generated =
	[ ("_includes_ensures_thm", ensures_to_GLEADSTO_lift_thm),
	  ("_includes_implies_thm", implies_to_GLEADSTO_lift_thm),
	  ("_includes_IMP_thm", IMP_to_GLEADSTO_lift_thm),
	  ("_transitive_thm", GLEADSTO_transitive_thm),
	  ("_transitive_thm_morph1", GLEADSTO_transitive_thm_morph1),
	  ("_LeftDisj_thm", GLEADSTO_LeftDisj_thm),
	  ("_LeftDisj_thm_morph1",  GLEADSTO_LeftDisj_thm_morph1),
	  ("_LeftDisj_thm_morph2", GLEADSTO_LeftDisj_thm_morph2),
	  ("_LeftDisj_thm_morph3", GLEADSTO_LeftDisj_thm_morph3),
	  ("_induction1_thm", GLEADSTO_induction1_thm),
	  ("_hasProperDomain_thm", GLEADSTO_hasProperDomain_thm),
	  ("_Subst_thm", GLEADSTO_Subst_thm ),
	  ("_Disj_thm", GLEADSTO_Disj_thm),
	  ("_RightDisj_thm", GLEADSTO_RightDisj_thm),
	  ("_SimpleDisjunction_thm", GLEADSTO_SimpleDisjunction_thm),
	  ("_Cancelation_thm", GLEADSTO_Cancelation_thm),
	  ("_BOUNDED_REACH_thm", GLEADSTO_BOUNDED_REACH_thm),
	  ("_BOUNDED_ALWAYS_REACH_thm", GLEADSTO_BOUNDED_ALWAYS_REACH_thm),
	  ("_PSP_thm_x", PSP_thm_x),
	  ("_Completion_with_stricterCond_thm",Completion_with_stricterCond_thm),
	  ("_Completion_thm_classical_with_stricterCond", Completion_thm_classical_with_stricterCond),
	  ("_Convergence_Conjunction_thm_with_stricterCond",Convergence_Conjunction_thm_with_stricterCond)

	] ;


(*---------------------------------------------------------------------------

WIZARD SECTION 0: Some general tactics, conversions, etc...

----------------------------------------------------------------------------*)

fun duplicate n x = if n<=0 then [] else x :: duplicate (n - 1) x ;

fun XRULE_ASSUM_TAC rule = RULE_ASSUM_TAC (liteLib.orelsef (rule,I)) ;
val UNDISCH_ALL_TAC = REPEAT (FIRST_ASSUM (UNDISCH_TAC o concl)) ;

val IMP_TO_AND_thm = 
	prove (--`p ==> q ==> r = p /\ q ==> r`--, PROVE_TAC []) ;

val NORMALIZE_FOR_MP_CONV
	=
	(REPEATC o CHANGED_CONV o DEPTH_CONV) RIGHT_IMP_FORALL_CONV
	THENC 
	(REPEATC o CHANGED_CONV o DEPTH_CONV) (REWR_CONV IMP_TO_AND_thm) ; 

(*---------------------------------------------------------------------------

WIZARD SECTION 1: Proving primitive properties.

In this section the wizard tries to prove the primitive properties of
your concrete UNITY operators. It has a pre-programmed tactic to do
this, but it may happen that your concrete UNITY operators are too
sophisticated for the wizard to prove. In this case you will have to
prove the properties on which the wizard fails yourself, which means
you should manually patch the proof codes in SUBSECTION 1.2.

Once all primitive properties are proven, the wizard will continue
with instantiating the rest of the (derived) UNITY laws. This will be
done in the next section, and should go without problem.

----------------------------------------------------------------------------*)


(*---------------------------------------------------------------------------

WIZARD SUBSECTION 1.1: preparing tactics, rules, utility functions etc.

----------------------------------------------------------------------------*)

val prop_defs1 = [
		isClosed_under_PredOPS_def,
		inDomain_def,
		includesIMP_def,
		isSubRelationOf_def,
		hasProperDomain_def,
		isConservative_def,
		implicitlyImplies_def ] ;

(*** default prover for implies' primitive properties ***)

val implies_REDUCE_TAC =
	REWRITE_TAC 
		(implies_def ::
		 implies_other_required_defs @
		 prop_defs1
		)
	THEN valid_PROVE_TAC ;

val prop_defs2 = [includesIMP_def,
		 inDomain_def,
		 isSubRelationOf_def,
		 hasProperDomain_def,
		 satisfiesUNLESS_AntiRefl_def, 
		 satisfiesUNLESS_Conj_def,
		 satisfiesUNLESS_Disj_def,
		 satisfiesUNLESS_Subst_def,
		 isConservative_def,
		 implicitlyImplies_def] ;

(*** default prover for unless' primitive properties ***)

val unless_REDUCE_TAC =
	REWRITE_TAC 
		([unless_def, implies_def] @
		 unless_other_required_defs @
		 prop_defs2
		)
	THEN valid_PROVE_TAC ;

val prop_defs3 = [includesIMP_def,
		 inDomain_def,
		 isSubRelationOf_def,
		 hasProperDomain_def,
	         satisfiesPSP_def,
		 isConservative_def,
		 implicitlyImplies_def ] ;

(*** default prover for ensures' primitive properties ***)

val ensures_REDUCE_TAC =
	REWRITE_TAC 
		([ensures_def, unless_def, implies_def] @
		 ensures_other_required_defs @
		 prop_defs3
		)
	THEN valid_PROVE_TAC ;


(*** a top level function to prove and save a primitive property ***)

exception UNITY_wizard_failure of unit ;

val line_marker = "***==============================***\n" ;

fun prove_primPROP 
	(op_name,
	 th_base_name,
	 property,
	 wizard_subsection_name,
	 tac)
	=
	let
	val th_name = op_name ^ th_base_name

	val fail_msg1 = "The UNITY wizard fails to prove property " ^
		   th_base_name ^ " of operator " ^ op_name ^ "." 

	val fail_msg2 = "You'll have to manually patch the corresponding " ^
			"proof in WIZARD (SUB)SECTION " ^ 
			wizard_subsection_name ^ "."
	in
	store_thm(th_name, property, tac)
	handle
	_ => raise UNITY_wizard_failure 
		(print "\n\n" ;
		 print line_marker ;
		 print fail_msg1 ;
		 print "\n\n" ;
		 print fail_msg2 ;
		 print "\n" ;
		 print line_marker ;
		 print "\n" )
	end; 

(*** a top level function to instantiate/generate and save a derived property ***)

fun mk_derivedPROP 
	op_name 
	th_base_name
	th_new_name
	th_source
	transformer
	wizard_subsection_name
	=
	let
	val th_name = op_name ^ th_base_name

	val th_new = transformer th_source

	val msg1 = "UNITY wizard fails to make the property " ^
		   th_base_name ^ " of the operator " ^ op_name ^ "."

	val msg2 = "Check the code in WIZARD SUBSECTION " ^
		   wizard_subsection_name ^ "."

	in

	save_thm(th_new_name, th_new)
	handle
	_ => raise UNITY_wizard_failure 
		(print "\n\n" ;
		 print line_marker ;
		 print msg1 ;
		 print "\n\n" ;
		 print msg2 ;
		 print "\n" ;
		 print line_marker ;
		 print "\n" )
	end ;

(*** 

A rule to specialize unity operators in a general law to concrete
operators.

***)

fun SPEC_UNITY thm
	=
	let

	fun SPEC_U thm =
		let
		val (bvar,_)      = dest_forall (concl thm)
		val (bvar_name,_) = dest_var bvar
		in
		case bvar_name of
		    "implies"  => ISPEC implies_object thm
		  | "unless"   => ISPEC unless_object thm
		  | "ensures"  => ISPEC ensures_object thm
		  | _	       => SPEC bvar thm
		end

	fun SPEC_UREC thm =
		if   is_forall (concl thm)
		then SPEC_UREC (SPEC_U thm)
		else thm

	in
	SPEC_UREC (GEN_ALL thm)
	end ;


(*** 

Conversions and rule to remove primitive properties assumptions from a
derived law.

***)

fun primPROP_REWR_CONV thm =
	let
	val thm1 = UNDISCH thm handle _ => thm
	val thm2 = prove(mk_eq(concl thm1, --`T`--), REWRITE_TAC [thm1])
	in
	REWR_CONV thm2 
	end ;

fun primPROPS_REWR_CONV [] = ALL_CONV
  | primPROPS_REWR_CONV (thm::the_rest) 
	=
	(REPEATC o CHANGED_CONV o DEPTH_CONV o primPROP_REWR_CONV) thm
	 THENC
	 (primPROPS_REWR_CONV the_rest) ;


fun REMOVE_primPROPS prim_props
	= 
	(CONV_RULE o primPROPS_REWR_CONV) prim_props ;

(*** 

A rule to clean the remaining T's that result after removing primitive
properties assumptions from a derived law.

***)

(* val CLEANING = CONV_RULE (SimpLib.SIMP_CONV std_ss []) ; *)

val CLEANING = REWRITE_RULE [] ;

(*** A rule to normalize the resulting law ***)

val NORMALIZING = CONV_RULE NORMALIZE_FOR_MP_CONV o DISCH_ALL ;

(*** 

Finally, composing them together, here is the rule to generate a
derived law.

***)

fun mk_derived_law prim_props gen_derived_law
	=
	(NORMALIZING o
	 CLEANING o
	 REMOVE_primPROPS prim_props o
	 REWRITE_RULE [SYM (SPEC_ALL leadsto_def)] o
	 SPEC_UNITY)
	 gen_derived_law ;

(*---------------------------------------------------------------------------

WIZARD SUBSECTION 1.2

UNITY primitive properties are listed here. You should prove them
first before you can inherit the laws from the general
UNITY. Pre-programmed tactics are given, but they may fail if your
concrete UNITY is too complicated for them. In this case you'll have
to prove them by hand. 

You should just modify the proof (tactic) section of the failing
properties. Let the naming of the properties unchanged so the wizard
can identify them and use them to instantiate the general UNITY.

Some primitive properties may not generally hold. In this the wizard
will allow you to add a condition specifying when a given property
does hold. So, if P is a property, if it only holds under condition C,
then specify this to the wizard:

	--`C ==> P`--

The wizard will then substitute the occurences of the condition P in
the laws with C.

----------------------------------------------------------------------------*)

(*---------------------------------------------------------------------------

WIZARD SUBSECTION 1.2.1: primitive properties of implies

----------------------------------------------------------------------------*)


val domain_closed_TT_lemma = 
	prove
	(--`isClosed_under_PredOPS dom ==> dom TT`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def]) ;

val domain_closed_FF_lemma = 
	prove
	(--`isClosed_under_PredOPS dom ==> dom FF`--,
	RW_TAC std_ss [
		isClosed_under_PredOPS_def,
		prove(--`FF = NOT TT`--, pred_eq_PROVE_TAC)
		]
	) ;

val domain_closed_NOT_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\ dom p ==> dom (NOT p)`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def]) ;

val domain_closed_AND_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\
	    dom p /\ dom q ==> dom (p AND q)`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def]) ;

val domain_closed_OR_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\
	    dom p /\ dom q ==> dom (p OR q)`--,
	RW_TAC std_ss [
		isClosed_under_PredOPS_def,
		prove(--`p OR q = NOT(NOT p AND NOT q)`--, pred_eq_PROVE_TAC)
		]
	) ;

val domain_closed_IMP_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\
	    dom p /\ dom q ==> dom (p IMP q)`--,
	RW_TAC std_ss [
		isClosed_under_PredOPS_def,
		prove(--`p IMP q = NOT(p AND NOT q)`--, pred_eq_PROVE_TAC)
		]
	) ;

val domain_closed_qAND_lemma1 =
	prove
	(--`isClosed_under_PredOPS dom /\
	    (!p. W p ==> dom p)
	    ==>
	    dom (!!p::W. p)`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def] ) ;


val domain_closed_qAND_lemma2 =
	prove
	(--`isClosed_under_PredOPS (dom:('a->bool)->bool) /\
	    (!i:'index. W i ==> dom (f i))
	    ==>
	    dom (!!i::W. f i)`--,
	STRIP_TAC
	THEN ONCE_REWRITE_TAC [
		prove(--`(!!i::W. f i) 
			 =
			 (!!p::(\p. (?i. W i /\ (p = f i))). p)`--,
		      pred_eq_PROVE_TAC)
		]
	THEN MATCH_MP_TAC domain_closed_qAND_lemma1
	THEN RW_TAC std_ss []
	THEN PROVE_TAC []) ;

val domain_closed_qOR_lemma =
	prove
	(--`isClosed_under_PredOPS (dom:('a->bool)->bool) /\
	    (!i:'index. W i ==> dom (f i))
	    ==>
	    dom (??i::W. f i)`--,
	STRIP_TAC
	THEN ONCE_REWRITE_TAC [
		prove(--`(??i::W. f i) 
			 =
			 NOT (!!i::W. (\i. NOT (f i)) i)`--,
		      pred_eq_PROVE_TAC)
		]
	THEN MATCH_MP_TAC domain_closed_NOT_lemma
	THEN ASM_REWRITE_TAC []
	THEN MATCH_MP_TAC (GEN_ALL domain_closed_qAND_lemma2)
	THEN RW_TAC std_ss []
	THEN PROVE_TAC [domain_closed_NOT_lemma]) ;


val domain_closed_qOR_lemma2 =
	prove
	(--`isClosed_under_PredOPS (dom:('a->bool)->bool) /\
	    (!p. W p ==> dom p)
	    ==>
	    dom (??p::W. p)`--,

	ONCE_REWRITE_TAC [
		prove(--`(??p::W. p) = (??p::W. (\x. x) p)`--,
		      pred_eq_PROVE_TAC)
		]
	THEN RW_TAC std_ss [ domain_closed_qOR_lemma]
	) ;

val domain_closedness_lemmas =
	[domain_closed_TT_lemma,
	 domain_closed_FF_lemma,
	 domain_closed_NOT_lemma,
	 domain_closed_AND_lemma,
	 domain_closed_OR_lemma,
	 domain_closed_IMP_lemma ] ;

val implies_domain_TAC 
	=
	FIRST (map MATCH_MP_TAC 
		(domain_closed_qOR_lemma :: domain_closedness_lemmas))
	THEN
	ASM_REWRITE_TAC [] 
	THEN REPEAT CONJ_TAC ;


val tmpI_domain_isClosed_under_PredOPS_thm 
	=
	prove_primPROP

	(implies_name,

	 "_domain_isClosed_under_PredOPS_thm",

	 --`isClosed_under_PredOPS A
	    ==>
	    isClosed_under_PredOPS (inDomain ^implies_object)`--,

	"1.2",

	REWRITE_TAC 
		(implies_def ::
		 implies_other_required_defs @
		 prop_defs1
		)
	THEN valid_decompose_TAC
	THEN RW_TAC std_ss []
	) ;


val tmpI_includesIMP_thm
	=
	prove_primPROP

	(implies_name,

	 "_includesIMP_thm",

	  --`includesIMP ^implies_object`--,

	"1.2",

	 implies_REDUCE_TAC ) ;


val tmpI_hasProperDomain_thm
	=
	prove_primPROP

	(implies_name,

	 "_hasProperDomain_thm",

	 --`hasProperDomain ^implies_object`--,

	"1.2",

	 implies_REDUCE_TAC ) ;


val tmpI_inDomain_thm
	=
	prove_primPROP

	(implies_name,

	 "_inDomain_thm",

	  --`A p ==> (inDomain ^implies_object) p`--,

	"1.2",

	 implies_REDUCE_TAC ) ;


(*---------------------------------------------------------------------------

WIZARD SUBSECTION 1.2.2: primitive properties of unless

----------------------------------------------------------------------------*)

val tmpU_hasProperDomain_thm
	=
	prove_primPROP

	(unless_name,

	 "_hasProperDomain_thm",

	  --`hasProperDomain ^unless_object`--,

	"1.2",

	 unless_REDUCE_TAC ) ;

(*** 

The following primitive property requires an additional condition INV
P J !! 

***)

val tmpU_includes_implies_thm
	=
	prove_primPROP

	(unless_name,

	 "_includes_implies_thm",

	  --`STABLE P J ==> isSubRelationOf ^implies_object ^unless_object`--,

	"1.2",

	 unless_REDUCE_TAC ) ;


val tmpU_satisfiesUNLESS_AntiRefl_thm
	=
	prove_primPROP

	(unless_name,

	 "_satisfiesUNLESS_AntiRefl_thm",

	  --`isClosed_under_PredOPS A /\ 
	     STABLE P J 
	     ==> 
	     satisfiesUNLESS_AntiRefl ^unless_object`--,

	"1.2",

	REWRITE_TAC [satisfiesUNLESS_AntiRefl_def, unless_def, inDomain_def]
	THEN RW_TAC std_ss []
	THENL
	[ implies_domain_TAC,
	  UNDISCH_ALL_TAC THEN unless_REDUCE_TAC ]
	) ;


val tmpU_satisfiesUNLESS_Conj_thm
	=
	prove_primPROP

	(unless_name,

	 "_satisfiesUNLESS_Conj_thm",

	  --`isClosed_under_PredOPS A
	     ==>
	     satisfiesUNLESS_Conj ^unless_object`--,

	"1.2",
	
	 REWRITE_TAC [ satisfiesUNLESS_Conj_def, unless_def]
	 THEN RW_TAC std_ss []
	 THENL
	 [ REPEAT implies_domain_TAC,
	   REPEAT implies_domain_TAC,
	   UNDISCH_ALL_TAC THEN unless_REDUCE_TAC ]
	) ;


val tmpU_satisfiesUNLESS_Disj_thm
	=
	prove_primPROP

	(unless_name,

	 "_satisfiesUNLESS_Disj_thm",

	  --`isClosed_under_PredOPS A
	     ==>
	     satisfiesUNLESS_Disj ^unless_object`--,

	"1.2",

	 REWRITE_TAC [ satisfiesUNLESS_Disj_def, unless_def]
	 THEN RW_TAC std_ss []
	 THENL
	 [ REPEAT implies_domain_TAC,
	   REPEAT implies_domain_TAC,
	   UNDISCH_ALL_TAC THEN unless_REDUCE_TAC ]
	) ;



val tmpU_satisfiesUNLESS_Subst_thm
	=
	prove_primPROP

	(unless_name,

	 "_satisfiesUNLESS_Subst_thm",

	  --`satisfiesUNLESS_Subst ^implies_object ^unless_object`--,

	"1.2",

	 unless_REDUCE_TAC ) ;


val tmpU_DOM_IN_implies_DOM_thm
	=
	prove_primPROP

	(unless_name,

	 "_DOM_IN_implies_DOM_thm",

	  --`(!p. inDomain (^unless_object) p ==> inDomain ^implies_object p)`--,

	"1.2",

	 unless_REDUCE_TAC ) ;

(*---------------------------------------------------------------------------

WIZARD SUBSECTION 1.2.3: primitive properties of ensures

----------------------------------------------------------------------------*)

val tmpE_hasProperDomain_thm
	=
	prove_primPROP

	(ensures_name,

	 "_hasProperDomain_thm",

	  --`hasProperDomain ^ensures_object`--,

	"1.2",

	 ensures_REDUCE_TAC ) ;


(*** 

The following primitive property requires an additional condition INV
P J !! 

***)

val tmpE_includes_implies_thm
	=
	prove_primPROP

	(ensures_name,

	 "_includes_implies_thm",

	  --`STABLE P J ==> isSubRelationOf ^implies_object ^ensures_object`--,

	"1.2",

	 ensures_REDUCE_TAC ) ;


val tmpE_DOM_IN_implies_DOM_thm
	=
	prove_primPROP

	(ensures_name,

	 "_DOM_IN_implies_DOM_thm",

	  --`(!p. inDomain (^ensures_object) p ==> inDomain ^implies_object p)`--,

	"1.2",

	 ensures_REDUCE_TAC ) ;


val tmpE_PSP_thm 
	= 
	prove_primPROP

	(ensures_name,
	
	"_PSP_thm",

	--`isClosed_under_PredOPS A
	   ==>
	   satisfiesPSP ^ensures_object ^unless_object`--,
	
	"1.2",


	 REWRITE_TAC [ satisfiesPSP_def, ensures_def, unless_def]
	 THEN RW_TAC std_ss []
	 THENL
	 [ REPEAT implies_domain_TAC,
	   REPEAT implies_domain_TAC,
	   UNDISCH_ALL_TAC THEN ensures_REDUCE_TAC,
	   UNDISCH_ALL_TAC THEN ensures_REDUCE_TAC ]
	) ;


val tmpE_implies_unless_thm
	= 
	prove_primPROP

	(ensures_name,
	
	"_implies_unless_thm",

	--`isSubRelationOf ^ensures_object ^unless_object`--,
	
	"1.2",

	ensures_REDUCE_TAC) ;



(*---------------------------------------------------------------------------

WIZARD SUBSECTION 1.2.4

The following lists all primitive properties of UNITY operators. They
all should be proven by now. In Section 2 of the wizard will
automatically derive the rest of the UNITY laws.

----------------------------------------------------------------------------*)


val UNITY_primitive_properties 
	=
	[ tmpI_includesIMP_thm,
	  tmpI_hasProperDomain_thm, 
	  tmpI_inDomain_thm,
	  tmpI_domain_isClosed_under_PredOPS_thm , 
	  tmpU_hasProperDomain_thm,
	  tmpU_includes_implies_thm, 
	  tmpU_satisfiesUNLESS_AntiRefl_thm,
	  tmpU_satisfiesUNLESS_Conj_thm, 
	  tmpU_satisfiesUNLESS_Disj_thm, 
	  tmpU_satisfiesUNLESS_Subst_thm,
	  tmpU_DOM_IN_implies_DOM_thm,
	  tmpE_hasProperDomain_thm,
	  tmpE_includes_implies_thm,
	  tmpE_DOM_IN_implies_DOM_thm,
	  tmpE_PSP_thm,
	  tmpE_implies_unless_thm
	] ;


(*---------------------------------------------------------------------------

WIZARD SECTION 2: generating derived laws.

----------------------------------------------------------------------------*)


(*---------------------------------------------------------------------------

WIZARD SUBSECTION 2.1

The following derived laws are obtained by simply unfolding some
definitions in the UNITY primitive properties.

----------------------------------------------------------------------------*)

fun expand_primPROP op_name defs_to_expand th_base_name
	=
	let
	
	val TRANSFORMER =
		SPEC_ALL o 
		CONV_RULE NORMALIZE_FOR_MP_CONV o
		REWRITE_RULE defs_to_expand

	in

	mk_derivedPROP 
		op_name 
		th_base_name 
		(op_name ^ th_base_name ^ "_x")
		(theorem (op_name ^ th_base_name))
		TRANSFORMER
		"2.1"
	end ;

(*** 

The following make unless' primitive properties in their expanded
form.

***)

val unless_primPROPS_x
	=
	map (expand_primPROP unless_name prop_defs2)
		["_hasProperDomain_thm",
		 "_includes_implies_thm", 
		 "_satisfiesUNLESS_AntiRefl_thm",
		 "_satisfiesUNLESS_Conj_thm", 
		 "_satisfiesUNLESS_Disj_thm", 
		 "_satisfiesUNLESS_Subst_thm"
		]
	;

(*** 

The following make ensures' primitive properties in their expanded
form.

***)

val ensures_primPROPS_x 
	=
	map (expand_primPROP ensures_name prop_defs3)
		["_hasProperDomain_thm",
		 "_includes_implies_thm",
		 "_PSP_thm",
	  	 "_implies_unless_thm" ]
	;




(*---------------------------------------------------------------------------

WIZARD SUBSECTION 2.2: derived properties of unless

----------------------------------------------------------------------------*)


fun mk_derived_prop op_name section_name  (th_base_name,th_source)
	=
	let
	val TRANSFORMER = mk_derived_law UNITY_primitive_properties
	in

	mk_derivedPROP 
		op_name 
		th_base_name 
		(op_name ^ th_base_name)
		th_source
		TRANSFORMER
		section_name 
	end ;


val unless_derived_properties
	=
	map  
	(mk_derived_prop unless_name "2.2")
	unless_laws_to_be_generated ;

(*---------------------------------------------------------------------------

WIZARD SUBSECTION 2.3: derived properties of ensures

None at the moment...

----------------------------------------------------------------------------*)

val ensures_derived_properties
	=
	map  
	(mk_derived_prop ensures_name "2.4")
	ensures_laws_to_be_generated ;


(*---------------------------------------------------------------------------

WIZARD SUBSECTION 2.4: derived properties of leadsto.

----------------------------------------------------------------------------*)

val leadsto_derived_properties
	=
	map  
	(mk_derived_prop leadsto_name "2.4")
	leadsto_laws_to_be_generated ;


(*---------------------------------------------------------------------------

WIZARD SECTION 3: Handling additional parameters...

----------------------------------------------------------------------------*)


(*---------------------------------------------------------------------------

WIZARD SECTION 3.1 : 

   Proving laws of the form: 

	UNITY_OPERATOR P J A p q ==> STABLE P J

----------------------------------------------------------------------------*)


(* What's your extra parameter ? *)

val J_par = --`J:^Pred`-- ;

(* What's the property about this extra parameter which you want to prove? *)

val J_prop = --`STABLE (P:^Prog)`-- ;

(* Now, prove that your concrete unless and ensures implies J_prop. *)

val tmpU_implies_stab_thm
	=
	prove_primPROP

	(unless_name,

	"_implies_stab_thm",

	--`implicitlyImplies (\^J_par. ^unless_object) ^J_prop`--,

	"3.1",
	
	REWRITE_TAC [implicitlyImplies_def]
	THEN BETA_TAC
	THEN unless_REDUCE_TAC) ;

val tmpE_implies_stab_thm
	=
	prove_primPROP

	(ensures_name,

	"_implies_stab_thm",

	--`implicitlyImplies (\^J_par. ^ensures_object)  ^J_prop`--,

	"3.1",

	REWRITE_TAC [implicitlyImplies_def]
	THEN BETA_TAC
	THEN ensures_REDUCE_TAC) ;


(* The following will prove that your leadsto also satisfies J_prop *)

val leadsto_implicitlyImplies_lemma =
	( BETA_RULE o
	  ISPEC (mk_abs(J_par,ensures_object)) o
	  GEN (--`ensures_ : ^Pred -> ^UOP`--) o
	  ISPEC J_prop o
	  GEN (--`property:'asm->bool`--)
	)
	GLEADSTO_implicitlyImplies_thm ;

val tmpLT_implies_stab_thm
	=
	prove_primPROP

	(leadsto_name,

	"_implies_stab_thm",

	--`implicitlyImplies (\^J_par. ^leadsto_object) ^J_prop`--,

	"3.1",

	REWRITE_TAC [implicitlyImplies_def]
	THEN BETA_TAC
	THEN REWRITE_TAC [leadsto_def]
	THEN REPEAT STRIP_TAC
	THEN MATCH_MP_TAC leadsto_implicitlyImplies_lemma
	THEN RW_TAC std_ss [tmpE_implies_stab_thm]
	) ;

(* 

The following exports the just proven properties J_prop, but they are
now expanded (the definition of implicitlyImplies)

*)

(** redefining expand_primPROP **)

fun expand_primPROP op_name defs_to_expand th_base_name
	=
	let
	
	val TRANSFORMER =
		SPEC_ALL o 
		CONV_RULE NORMALIZE_FOR_MP_CONV o
		BETA_RULE o
		REWRITE_RULE defs_to_expand

	in

	mk_derivedPROP 
		op_name 
		th_base_name 
		(op_name ^ th_base_name ^ "_x")
		(theorem (op_name ^ th_base_name))
		TRANSFORMER
		"2.1"
	end ;


val _ = expand_primPROP unless_name [implicitlyImplies_def] "_implies_stab_thm" ;
val _ = expand_primPROP ensures_name [implicitlyImplies_def] "_implies_stab_thm" ;
val _ = expand_primPROP leadsto_name [implicitlyImplies_def] "_implies_stab_thm" ;

(* 

The following will prove that the derived operators COMPLETES and
CONVERGES also satisfies J_prop. 

*)

val COMPLETES_implies_stab_thm_x
	=
	store_thm
	("COMPLETES_implies_stab_thm_x",
	--`COMPLETES (^leadsto_object) (^unless_object) p q b 
	   ==>
	   ^J_prop ^J_par`--,
	PROVE_TAC [COMPLETES_def,  theorem (leadsto_name ^ "_implies_stab_thm_x") ]
	) ;

val CONVERGES_implies_stab_thm_x
	=
	store_thm
	("CONVERGES_implies_stab_thm_x",
	--`CONVERGES (^leadsto_object) (^unless_object) p q 
	   ==>
	   ^J_prop ^J_par`--,
	PROVE_TAC [CONVERGES_def, COMPLETES_implies_stab_thm_x]
	) ;


(*---------------------------------------------------------------------------

WIZARD SECTION 3.2.a : conservative composition/extension

   Proving laws of the form: 

	UNITY_OPERATOR P J A p q 
	==>
	UNITY_OPERATOR P (J' AND J) A p q 

----------------------------------------------------------------------------*)

(* define youe composition/extension operator *)

val J_extend = --`$AND (J':^Pred)`-- ;

(* 

Now you must prove that your unless and ensures preserve their
behavior when their J_par is extended with J_extend.

*)

val tmpU_is_J_Conservative_thm
	=
	prove_primPROP

	(unless_name,

	"_is_J_Conservative_thm",

	--`STABLE P J' 
	   ==> 
	   isConservative (\^J_par. ^unless_object) ^J_extend`--,

	"3.2",

	REWRITE_TAC [isConservative_def]
	THEN BETA_TAC
	THEN unless_REDUCE_TAC) ;


val tmpE_is_J_Conservative_thm
	=
	prove_primPROP

	(ensures_name,

	"_is_J_Conservative_thm",

	--`STABLE P J' 
	   ==> 
	   isConservative (\^J_par. ^ensures_object) ^J_extend`--,

	"3.2",

	REWRITE_TAC [isConservative_def]
	THEN BETA_TAC
	THEN ensures_REDUCE_TAC) ;

val _ = expand_primPROP unless_name  [isConservative_def] "_is_J_Conservative_thm" ;
val _ = expand_primPROP ensures_name [isConservative_def] "_is_J_Conservative_thm" ;


(* 

The following will prove that your leadsto is preserved by J_extend.

*)

val leadsto_conservative_lemma =
	( REWRITE_RULE [SYM (SPEC_ALL leadsto_def)] o
	  BETA_RULE o
	  ISPEC  (mk_abs(J_par,ensures_object)) o
	  GEN (--`ensures_ : ^Pred -> ^UOP`--) o
	  ISPEC J_extend o
	  GEN (--`extend:'asm->'asm`--)
	)
	GLEADSTO_conservative_extension_thm ;

val tmpLT_is_J_Conservative_thm_x
	=
	store_thm
	(leadsto_name ^ "_is_J_Conservative_thm_x",
	--`STABLE P J' /\ 
	   ^leadsto_object p q 
	   ==>
	   LEADSTO P (J' AND J) A p q`--,

	REPEAT STRIP_TAC
	THEN MATCH_MP_TAC leadsto_conservative_lemma 
	THEN RW_TAC std_ss [tmpE_is_J_Conservative_thm]
	) ;

(* 

The following will prove that the derived operators COMPLETES and
CONVERGES are also preserved by J_extend.

*)


val COMPLETES_is_J_Conservative_thm_x
	=
	store_thm
	("COMPLETES_is_J_Conservative_thm_x",
	--`STABLE P J' /\ 
	   COMPLETES (LEADSTO P J A) (UNLESS P J A) p q b
	   ==>
	   COMPLETES (LEADSTO P (J' AND J) A) 
		     (UNLESS  P (J' AND J) A)
		p q b`--,

	REWRITE_TAC [COMPLETES_def]
	THEN REPEAT STRIP_TAC
	THEN PROVE_TAC [theorem (leadsto_name ^ "_is_J_Conservative_thm_x"),
		   	theorem (unless_name ^ "_is_J_Conservative_thm_x")  ]

	) ;



val CONVERGES_is_J_Conservative_thm_x
	=
	store_thm
	("CONVERGES_is_J_Conservative_thm_x",
	--`STABLE P J' /\ 
	   CONVERGES (LEADSTO P J A) (UNLESS P J A) p q 
	   ==>
	   CONVERGES (LEADSTO P (J' AND J) A) 
		     (UNLESS  P (J' AND J) A)
		p q`--,	   

	PROVE_TAC [CONVERGES_def,  
		   COMPLETES_is_J_Conservative_thm_x]
	) ;


(*---------------------------------------------------------------------------

WIZARD SECTION 3.2.b : conservative composition/extension

   Proving laws of the form: 

	UNITY_OPERATOR P J A p q 
	==>
	UNITY_OPERATOR P (J' AND J) A p q 

----------------------------------------------------------------------------*)

(* define youe composition/extension operator *)

val A_extend = --`$OR (B:^Pred->bool)`-- ;

(* 

Now you must prove that your unless and ensures preserve their
behavior when their J_par is extended with J_extend.

*)

val tmpU_is_A_Conservative_thm
	=
	prove_primPROP

	(unless_name,

	"_is_A_Conservative_thm",

	--`isConservative ^(rator unless_object) ^A_extend`--,

	"3.2",

	unless_REDUCE_TAC) ;


val tmpE_is_A_Conservative_thm
	=
	prove_primPROP

	(ensures_name,

	"_is_A_Conservative_thm",

	--`isConservative ^(rator ensures_object) ^A_extend`--,

	"3.2",

	ensures_REDUCE_TAC) ;

val _ = expand_primPROP unless_name  [isConservative_def] "_is_A_Conservative_thm" ;
val _ = expand_primPROP ensures_name [isConservative_def] "_is_A_Conservative_thm" ;


(* 

The following will prove that your leadsto is preserved by J_extend.

*)

val leadsto_conservative_lemma =
	( REWRITE_RULE [SYM (SPEC_ALL leadsto_def)] o
	  BETA_RULE o
	  ISPEC (rator ensures_object) o
	  GEN (--`ensures_ : (^Pred->bool) -> ^UOP`--) o
	  ISPEC A_extend o
	  GEN (--`extend:'asm->'asm`--)
	)
	GLEADSTO_conservative_extension_thm ;

val tmpLT_is_A_Conservative_thm_x
	=
	store_thm
	(leadsto_name ^ "_is_A_Conservative_thm_x",
	--`^leadsto_object p q 
	   ==>
	   LEADSTO P J (B OR A) p q`--,

	REPEAT STRIP_TAC
	THEN MATCH_MP_TAC leadsto_conservative_lemma 
	THEN RW_TAC std_ss [tmpE_is_A_Conservative_thm]
	) ;

(* 

The following will prove that the derived operators COMPLETES and
CONVERGES are also preserved by J_extend.

*)


val COMPLETES_is_A_Conservative_thm_x
	=
	store_thm
	("COMPLETES_is_A_Conservative_thm_x",
	--`COMPLETES (LEADSTO P J A) (UNLESS P J A) p q b
	   ==>
	   COMPLETES (LEADSTO P J (B OR A)) 
		     (UNLESS  P J (B OR A))
		p q b`--,

	REWRITE_TAC [COMPLETES_def]
	THEN REPEAT STRIP_TAC
	THEN PROVE_TAC [theorem (leadsto_name ^ "_is_A_Conservative_thm_x"),
		   	theorem (unless_name ^ "_is_A_Conservative_thm_x")  ]

	) ;



val CONVERGES_is_A_Conservative_thm_x
	=
	store_thm
	("CONVERGES_is_A_Conservative_thm_x",
	--`CONVERGES (LEADSTO P J A) (UNLESS P J A) p q 
	   ==>
	   CONVERGES (LEADSTO P J (B OR A)) 
		     (UNLESS  P J (B OR A))
		p q`--,	   

	PROVE_TAC [CONVERGES_def,  
		   COMPLETES_is_A_Conservative_thm_x]
	) ;


(*---------------------------------------------------------------------------

USER SECTION 3: cleaning....

The assumptions of the theorems generated by the wizard may still have
some residual terms which are superflous (implied by other
assumptions). You may want to write a small script to clean this kind
of terms.

Redefine mk_derived_law by adding your own cleaning script. Then rerun
the definition of mk_derived_prop, then call this function once more
to re-generate the derived properties.

----------------------------------------------------------------------------*)

val shiftright_J_thm 
	=
	 PROVE [] `INV P (J:^Pred) /\ q /\ r = (INV P J /\ q) /\ r`;

val tmpLT_clean_Stab_lemma 
	= 
	prove(--`STABLE P J /\ ^leadsto_object p q = ^leadsto_object p q`--,
		PROVE_TAC [theorem (leadsto_name ^ "_implies_stab_thm_x")]) ;

val tmpCOMPLETES_clean_Stab_lemma 
	= 
	prove(--`STABLE P J /\ 
		 COMPLETES (^leadsto_object) (^unless_object) p q b
		 = 
		 COMPLETES (^leadsto_object) (^unless_object) p q b`--,
		PROVE_TAC [COMPLETES_implies_stab_thm_x]) ;

val tmpCONVERGES_clean_Stab_lemma 
	= 
	prove(--`STABLE P J /\ 
		 CONVERGES (^leadsto_object) (^unless_object) p q 
		 = 
		 CONVERGES (^leadsto_object) (^unless_object) p q`--,
		PROVE_TAC [CONVERGES_implies_stab_thm_x]) ;

(** my cleaning script **)

val USER_CLEANING_RULE = 
	CONV_RULE (SimpLib.SIMP_CONV std_ss [SYM shiftright_J_thm])
	o
	CONV_RULE (SimpLib.SIMP_CONV std_ss 
		[shiftright_J_thm,
		 tmpLT_clean_Stab_lemma,
		 tmpCOMPLETES_clean_Stab_lemma,
		 tmpCONVERGES_clean_Stab_lemma ]
		)
	;

(*---------------------------------------------------------------------------

   Redefining mk_derived_law prim_props by adding the
   USER_CLEANING_RULE as its final step.

----------------------------------------------------------------------------*)

fun mk_derived_law prim_props gen_derived_law
	=
	(USER_CLEANING_RULE o
	 NORMALIZING o
	 CLEANING o
	 REMOVE_primPROPS prim_props o
	 REWRITE_RULE [SYM (SPEC_ALL leadsto_def)] o
	 SPEC_UNITY)
	 gen_derived_law ;


(*---------------------------------------------------------------------------
  Re-executing this function...
----------------------------------------------------------------------------*)

fun mk_derived_prop op_name section_name  (th_base_name,th_source)
	=
	let
	val TRANSFORMER = mk_derived_law UNITY_primitive_properties
	in

	mk_derivedPROP 
		op_name 
		th_base_name 
		(op_name ^ th_base_name)
		th_source
		TRANSFORMER
		section_name 
	end ;

(*---------------------------------------------------------------------------
  
  Now regenerating the leadsto proprties...

----------------------------------------------------------------------------*)

val leadsto_derived_properties
	=
	map  
	(mk_derived_prop leadsto_name "2.4")
	leadsto_laws_to_be_generated ;



val _ = export_theory () ;

