(*-----------------------------------------------------------------------------------
  
   Written by Wishnu Prasetya
   Dept. of Computer Science,
   Utrecht University
   Email : wishnu@cs.uu.nl

   Last update 
	March 2002, Taupo 6
	Feb 2002, Taupo 6
	Apr 99
-----------------------------------------------------------------------------------*)           
open Parse HolKernel basicHol90Lib ;
infix THEN THENL THENC ORELSE ORELSEC THEN_TCL ORELSE_TCL ## |-> ;
infixr --> ;


val _ = new_theory "generalUNITY" ;


(*
load "bossLib" ;
load "predTheory" ;
load "predLib" ;
load "combinTheory" ;
*)

(*
load "res_quanTheory" ;
load "uwfTheory" ;
load "btacs" ;
*)

open bossLib ;
open predTheory ;
open predLib ;
open PSyntax ;

(*
open res_quanTheory ;
open uwfTheory ;
open urelTheory ;
open btacs ;
*)

val _ = associate_restriction ("!!", "RES_qAND") ;
val _ = associate_restriction ("??", "RES_qOR") ;


(*-----------------------------------------------------------------------------------

Some general tactics, conversions, etc...

------------------------------------------------------------------------------------*)

fun XRULE_ASSUM_TAC rule = RULE_ASSUM_TAC (liteLib.orelsef (rule,I)) ;
val UNDISCH_ALL_TAC = REPEAT (FIRST_ASSUM (UNDISCH_TAC o concl)) ;

val IMP_TO_AND_thm = 
	prove (--`p ==> q ==> r = p /\ q ==> r`--, PROVE_TAC []) ;

val NORMALIZE_FOR_MP_CONV
	=
	(REPEATC o CHANGED_CONV o DEPTH_CONV) RIGHT_IMP_FORALL_CONV
	THENC 
	(REPEATC o CHANGED_CONV o DEPTH_CONV) (REWR_CONV IMP_TO_AND_thm) ; 

(*-----------------------------------------------------------------------------------

Some ML names for type abbreviations... 

------------------------------------------------------------------------------------*)

val Pred  = ty_antiq (Type `:('a->bool)`) ;
val UOP   = ty_antiq (Type `:('a->bool)->('a->bool)->bool`) ;

(*-----------------------------------------------------------------------------------

Some general definitions, not directly related to UNITY stuff.

------------------------------------------------------------------------------------*)

val isSubRelationOf_def 
	= 
	Define
	`isSubRelationOf U V = !x y. U x y ==> V x y` ;

(*-----------------------------------------------------------------------------------

  Define the closedness of a domain under standard predicate operators.
  When a domain is closed, then it is a sub-algebra, thus satisfying
  the same algebraic properties of ordinary predicates.

------------------------------------------------------------------------------------*)

val isClosed_under_PredOPS_def = Define
	`isClosed_under_PredOPS dom
	 =
	 dom TT /\ 
	 (!p. dom p ==> dom (NOT p)) /\
	 (!p q. dom p /\ dom q ==> dom (p AND q)) /\
	 (!W. (!p. W p ==> dom p) ==> dom (!!p::W. p))` ;

val domain_closed_TT_lemma = 
	prove
	(--`isClosed_under_PredOPS dom ==> dom TT`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def]) ;

val domain_closed_FF_lemma = 
	prove
	(--`isClosed_under_PredOPS dom ==> dom FF`--,
	RW_TAC std_ss [
		isClosed_under_PredOPS_def,
		prove(--`FF = NOT TT`--, pred_eq_PROVE_TAC)
		]
	) ;

val domain_closed_NOT_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\ dom p ==> dom (NOT p)`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def]) ;

val domain_closed_AND_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\
	    dom p /\ dom q ==> dom (p AND q)`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def]) ;

val domain_closed_OR_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\
	    dom p /\ dom q ==> dom (p OR q)`--,
	RW_TAC std_ss [
		isClosed_under_PredOPS_def,
		prove(--`p OR q = NOT(NOT p AND NOT q)`--, pred_eq_PROVE_TAC)
		]
	) ;

val domain_closed_IMP_lemma = 
	prove
	(--`isClosed_under_PredOPS dom /\
	    dom p /\ dom q ==> dom (p IMP q)`--,
	RW_TAC std_ss [
		isClosed_under_PredOPS_def,
		prove(--`p IMP q = NOT(p AND NOT q)`--, pred_eq_PROVE_TAC)
		]
	) ;

val domain_closed_qAND_lemma1 =
	prove
	(--`isClosed_under_PredOPS dom /\
	    (!p. W p ==> dom p)
	    ==>
	    dom (!!p::W. p)`--,
	RW_TAC std_ss [isClosed_under_PredOPS_def] ) ;


val domain_closed_qAND_lemma2 =
	prove
	(--`isClosed_under_PredOPS (dom:('a->bool)->bool) /\
	    (!i:'index. W i ==> dom (f i))
	    ==>
	    dom (!!i::W. f i)`--,
	STRIP_TAC
	THEN ONCE_REWRITE_TAC [
		prove(--`(!!i::W. f i) 
			 =
			 (!!p::(\p. (?i. W i /\ (p = f i))). p)`--,
		      pred_eq_PROVE_TAC)
		]
	THEN MATCH_MP_TAC domain_closed_qAND_lemma1
	THEN RW_TAC std_ss []
	THEN PROVE_TAC []) ;

val domain_closed_qOR_lemma =
	prove
	(--`isClosed_under_PredOPS (dom:('a->bool)->bool) /\
	    (!i:'index. W i ==> dom (f i))
	    ==>
	    dom (??i::W. f i)`--,
	STRIP_TAC
	THEN ONCE_REWRITE_TAC [
		prove(--`(??i::W. f i) 
			 =
			 NOT (!!i::W. (\i. NOT (f i)) i)`--,
		      pred_eq_PROVE_TAC)
		]
	THEN MATCH_MP_TAC domain_closed_NOT_lemma
	THEN ASM_REWRITE_TAC []
	THEN MATCH_MP_TAC (GEN_ALL domain_closed_qAND_lemma2)
	THEN RW_TAC std_ss []
	THEN PROVE_TAC [domain_closed_NOT_lemma]) ;


val domain_closed_qOR_lemma2 =
	prove
	(--`isClosed_under_PredOPS (dom:('a->bool)->bool) /\
	    (!p. W p ==> dom p)
	    ==>
	    dom (??p::W. p)`--,

	ONCE_REWRITE_TAC [
		prove(--`(??p::W. p) = (??p::W. (\x. x) p)`--,
		      pred_eq_PROVE_TAC)
		]
	THEN RW_TAC std_ss [ domain_closed_qOR_lemma]
	) ;

val domain_closedness_lemmas =
	[domain_closed_TT_lemma,
	 domain_closed_FF_lemma,
	 domain_closed_NOT_lemma,
	 domain_closed_AND_lemma,
	 domain_closed_OR_lemma,
	 domain_closed_IMP_lemma ] ;

(*-----------------------------------------------------------------------------------

The (implicit) domain of UNITY operators can typically be
characterized as follows:

------------------------------------------------------------------------------------*)

val inDomain_def = Define `inDomain U p = U p p` ;


(*-----------------------------------------------------------------------------------

Expected property of the domains of typical UNITY operators:

------------------------------------------------------------------------------------*)

val hasProperDomain_def = Define
	`hasProperDomain U
	 =
	 !p q. U p q ==> inDomain U p /\ inDomain U q` ;

(*-----------------------------------------------------------------------------------

  Many UNITY operators includes (lifts) the ordinary IMPlication. This 
  property is now weakened by requiring inclusion only when the arguments
  are 'in the domain' of the operator.

------------------------------------------------------------------------------------*)

val includesIMP_def = Define
	`includesIMP U
	 =
	 !p q. inDomain U p /\ 
	       inDomain U q /\ 
	       valid (p IMP q) 
	       ==> U p q` ;


(*-----------------------------------------------------------------------------------

  Some basic algebraic properties of typical in an UNLESS-like operator:

	* Anti reflexifity
	* The general disjunction and conjunction property
	* Substitution axiom

------------------------------------------------------------------------------------*)

val satisfiesUNLESS_AntiRefl = Define
	`satisfiesUNLESS_AntiRefl unless
	 =
	 !p. inDomain unless p ==> unless p (NOT p)` ;

val satisfiesUNLESS_Conj_def = Define
	`satisfiesUNLESS_Conj unless
	 =
	 !p1 q1 p2 q2.
		unless p1 q1 /\ unless p2 q2 
		==>
		unless (p1 AND p2) ((q1 AND p2) OR (q2 AND p1) OR (q1 AND q2))` ;

val satisfiesUNLESS_Disj_def = Define
	`satisfiesUNLESS_Disj unless
	 =
	 !p1 q1 p2 q2.
		unless p1 q1 /\ unless p2 q2 
		==>
		unless (p1 OR p2) 
		       ((q1 AND NOT p2) OR (q2 AND NOT p1) OR (q1 AND q2))` ;

(*-----------------------------------------------------------------------------------

Here is the formulation of substitution law (for unless). Note that no
unsoundness is introduced since 'implies' is a parameter!! 

For example, in the case of the classical UNITY 'implies' will be the
ordinary tautology implication. In the case of Sanders' UNITY then you
may strengthen this with invariant information.

------------------------------------------------------------------------------------*)


val satisfiesUNLESS_Subst_def = Define
	`satisfiesUNLESS_Subst implies unless
	 =
	 !p q a b. unless p q /\ 
		   implies p a /\ implies a p /\
		   implies q b
		   ==>
		   unless a b` ;

val UNLESS_logic_defs = 
	[includesIMP_def,
	 satisfiesUNLESS_Subst_def,	
	 satisfiesUNLESS_Conj_def,
	 satisfiesUNLESS_Disj_def,
	 hasProperDomain_def ] ;

(*-----------------------------------------------------------------------------------

Derived laws for unless-like operators:

	*inclusion of the noremal IMP
	*reflexivity
	*post weakening
	*simple conjunction and disjunction

------------------------------------------------------------------------------------*)


val implies_to_unless_thm = store_thm
	("implies_to_unless_thm",
	--`isSubRelationOf implies unless /\
	   implies p q
	   ==>
	   unless p q`--,
	PROVE_TAC [isSubRelationOf_def]) ;

val IMP_to_unless_thm = store_thm
	("IMP_to_unless_thm",
	--`includesIMP implies /\
	   isSubRelationOf implies unless /\
	   inDomain implies p /\ inDomain implies q /\
	   valid (p IMP q)
	   ==>
	   unless p q`--,
	PROVE_TAC [isSubRelationOf_def, includesIMP_def]) ;


val unless_REFL_thm = store_thm
	("unless_REFL_thm",
	--`inDomain unless p
	   ==>
	   unless p p`--,
	PROVE_TAC [inDomain_def]) ;

val unless_post_weakening_thm
	=
	store_thm
	("unless_post_weakening_thm",
	 --`includesIMP implies /\
	    (!p. inDomain unless p ==> inDomain implies p) /\
	    hasProperDomain unless /\
	    satisfiesUNLESS_Subst implies (unless:^UOP) /\
	    unless p q /\
	    implies q r
	    ==>
	    unless p r`--,
	RW_TAC std_ss UNLESS_logic_defs
	THEN RES_TAC THEN RES_TAC
	THEN PROVE_TAC [prove(--`valid (p IMP p)`--, valid_PROVE_TAC)]) ;


val unless_simple_conjunction_thm
	=
	store_thm
	("unless_simple_conjunction_thm",
	--`includesIMP implies /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   hasProperDomain unless /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   unless p1 q1 /\
	   unless p2 q2
	   ==>
	   unless (p1 AND p2) (q1 OR q2)`--,

	RW_TAC std_ss [satisfiesUNLESS_Conj_def]
	THEN MATCH_MP_TAC (GEN_ALL unless_post_weakening_thm)
	THEN EXISTS_TAC (--`(q1 AND p2) OR (q2 AND p1) OR (q1 AND (q2:^Pred))`--)
 	THEN EXISTS_TAC (--`implies:^UOP`--)
	THEN RW_TAC std_ss []
	THEN XRULE_ASSUM_TAC (REWRITE_RULE [includesIMP_def])
	THEN FIRST_ASSUM MATCH_MP_TAC 
	THEN REPEAT CONJ_TAC
	THEN TRY valid_PROVE_TAC
	THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
	THEN PROVE_TAC domain_closedness_lemmas
	) ;

val unless_simple_disjunction_thm
	=
	store_thm
	("unless_simple_disjunction_thm",
	--`includesIMP implies /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   hasProperDomain unless /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Disj unless /\
	   unless p1 q1 /\
	   unless p2 q2
	   ==>
	   unless (p1 OR p2) (q1 OR q2)`--,

	RW_TAC std_ss [satisfiesUNLESS_Disj_def]
	THEN MATCH_MP_TAC (GEN_ALL unless_post_weakening_thm)
	THEN EXISTS_TAC (--`(q1 AND NOT p2) OR (q2 AND NOT p1) OR (q1 AND (q2:^Pred))`--)
 	THEN EXISTS_TAC (--`implies:^UOP`--)
	THEN RW_TAC std_ss []
	THEN XRULE_ASSUM_TAC (REWRITE_RULE [includesIMP_def])
	THEN FIRST_ASSUM MATCH_MP_TAC 
	THEN REPEAT CONJ_TAC
	THEN TRY valid_PROVE_TAC
	THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
	THEN PROVE_TAC domain_closedness_lemmas
	) ;

(*-----------------------------------------------------------------------------------

Leadsto like operators are typically transitive and
disjunctive. Typically, they are left and right disjunctive, though
because the typically also right monotonic it suffices to require them
to be left disjunctive.

Here we define: different notions of transitivuty and disjunctivity,
and some other things which are helpful for later.

Note that isLeftDisj is the weakest among the disjunctivity notions we
give. This is what is later assumed by our LEADSTO like operators.

------------------------------------------------------------------------------------*)

val isLeftDisj_def  
	=  
	Define
	`isLeftDisj leadsto
	 =  
	 (!W q. (?p. W p) /\ (!p. W p ==> leadsto p q)
	       ==> 
	      leadsto (??p::W. p) q)` ;

val isLeftDisjunctive_def  
	=  
	Define
	`isLeftDisjunctive W leadsto
	 =  
	 !f q.  (!i. W i ==> leadsto (f i) q) ==> leadsto (??i::W. f i) q` ;


val isDisjunctive_def 
	= 
	Define
	`isDisjunctive W leadsto 
	=  
	!f g.  (!i. W i ==> leadsto (f i) (g i))
                ==> 
                leadsto (??i::W. f i) (??i::W. g i)` ;

val isTransitive_def  
	= 
	Define
	`isTransitive leadsto 
	 = 
	 !p q r. leadsto p q /\ leadsto q r ==> leadsto p r` ; 

val isTransitive2_def  
	= 
	Define
	`isTransitive2 ens leadsto 
	 = 
	 !p q r. ens p q  /\  leadsto q r ==> leadsto p r` ; 


val rINTER_def
	=
	Define
	`rINTER U V = (\x y. U x y /\ V x y)` ;


(*-----------------------------------------------------------------------------------

Relation between different disjunctivity properties. In particular,
these theorems give sufficient conditions for LEADSTO-like operators
to be generally disjunctive.

If we have general disjunctivity, we can prove laws like cancelation
and bounded progres.

------------------------------------------------------------------------------------*)

val LeftDisj_imp_LeftDisjunction_thm
	=
	store_thm
	("LeftDisj_imp_LeftDisjunction_thm",
	--`isLeftDisj leadsto /\ W (i:'index)
	   ==>
	   isLeftDisjunctive W leadsto`--,
	RW_TAC std_ss [isLeftDisj_def, isLeftDisjunctive_def]
	THEN ONCE_REWRITE_TAC [
		prove(--`RES_qOR W f 
			 = 
			 (??p::(\p. ?i. W i /\ (p = f i)). p)`--,
			 pred_eq_PROVE_TAC) ]
	THEN FIRST_ASSUM MATCH_MP_TAC
	THEN RW_TAC std_ss []
	THEN PROVE_TAC [] 
	) ;

val LeftDisj_imp_Disjunction_thm
	=
	store_thm
	("LeftDisj_imp_Disjunction_thm",
	--`includesIMP implies /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   isSubRelationOf implies leadsto /\
	   (!p. inDomain leadsto p ==> inDomain implies p) /\
	   hasProperDomain leadsto /\
	   isLeftDisj (leadsto:^UOP) /\
	   isTransitive leadsto /\
	   W (i:'index) 
	   ==>
	   isDisjunctive W leadsto`--,

	REWRITE_TAC [isDisjunctive_def,
		     isTransitive_def]
	THEN REPEAT STRIP_TAC
	THEN (MATCH_MP_TAC  o 
	      (CONV_RULE NORMALIZE_FOR_MP_CONV)   o 
	      REWRITE_RULE [isLeftDisjunctive_def]  o 
	      GEN_ALL)
	      LeftDisj_imp_LeftDisjunction_thm
	THEN EXISTS_TAC (--`i:'index`--)
	THEN ASM_REWRITE_TAC []
	THEN REPEAT STRIP_TAC
	THEN SUBGOAL_THEN
	     (--`(leadsto:^UOP) (g i') (??i:'index::W. g i)`--)
	     ASSUME_TAC
	THENL [ ALL_TAC , PROVE_TAC []]
	THEN
	(* cont. with proving the lemma *)
	SUBGOAL_THEN
	     (--`(implies:^UOP) (g i') (??i:'index::W. g i)`--)
	     ASSUME_TAC
	THENL [ ALL_TAC , PROVE_TAC [isSubRelationOf_def]]
	THEN
	(* cont. with proving the lemma *)
	XRULE_ASSUM_TAC (REWRITE_RULE [includesIMP_def])
	THEN FIRST_ASSUM MATCH_MP_TAC
	THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
	THEN REPEAT CONJ_TAC
	THEN TRY valid_PROVE_TAC
	THEN MATCH_MP_TAC (GEN_ALL domain_closed_qOR_lemma)
	THEN PROVE_TAC []
	) ;

(*** this trivial lemma is handy for some proofs later ***)

val LeftDisjuction_lemma = prove
	(--`isDisjunctive W leadsto /\ W (i:'index) 
	   ==> 
	   isLeftDisjunctive W leadsto`--,
	let
	val lemma = prove
		(--`W (i:'index) ==> ((q:'b->bool) = (??i::W. (\i. q) i))`--,
		 STRIP_TAC THEN 
		 pred_eq_PROVE_TAC)
	in
	REWRITE_TAC [isDisjunctive_def, isLeftDisjunctive_def]
	THEN REPEAT STRIP_TAC
	THEN SUBST1_TAC (UNDISCH lemma)
	THEN FIRST_ASSUM MATCH_MP_TAC
	THEN RW_TAC std_ss [] 
	end
	) ;	


(*-----------------------------------------------------------------------------------

Some laws for LEADSTO-like operators which are derivable from the
transitivity and disjunctivity properties:

	*The subsituion law
	 
	Note that the law is actually the combination of
	pre-strengthening and post-weakening.

	*Right disjunctivity law
	*Simple disjunction law
	*Cancelation law

------------------------------------------------------------------------------------*)

val leadsto_Subst_thm 
	=
	store_thm
	("leadsto_Subst_thm",
	--`isSubRelationOf implies leadsto /\
	   isTransitive leadsto /\
	   implies a p /\ implies q b /\
	   leadsto p q
	   ==>
	   leadsto a b`--,
	
	RW_TAC std_ss [isSubRelationOf_def, isTransitive_def]
	THEN PROVE_TAC []) ;
	   

val leadsto_GeneralRightDisjunction_thm
	=
	store_thm 
  	("leadsto_GeneralRightDisjunction_thm",
	--`isDisjunctive W leadsto /\ 
	   W i /\
	   (!i. W i ==> leadsto p (f i))
           ==>
           leadsto p (??i::W. f i)`--,
	let
	val lemma = prove
		(--`W i ==> (p = (??i::W. (\i. p) i))`--,
		 STRIP_TAC THEN 
		 pred_eq_PROVE_TAC)
	in
	REWRITE_TAC [isDisjunctive_def]
	THEN STRIP_TAC
	THEN SUBST1_TAC (UNDISCH lemma)
	THEN FIRST_ASSUM MATCH_MP_TAC
	THEN RW_TAC std_ss [] THEN PROVE_TAC []
	end
	) ;


val leadsto_SimpleDisjunction_thm = store_thm 
  ("leadsto_SimpleDisjunction_thm",
    --`isDisjunctive (TT:bool->bool) leadsto /\
       leadsto p q /\ leadsto r s
       ==>
       leadsto (p OR r) (q OR s)`--,
	let
	val lemma = prove
		(--`p OR q = (??b::TT. if b then p else q)`--,
		pred_eq_decompose_TAC
		THEN RW_TAC std_ss []
		THEN EQ_TAC
		THEN TRY (PROVE_TAC [])
		THEN RW_TAC std_ss []
		THENL [EXISTS_TAC (--`T`--) THEN RW_TAC std_ss [],
		       EXISTS_TAC (--`F`--) THEN RW_TAC std_ss []]
		) 
	in
	REWRITE_TAC [TT_def, isDisjunctive_def, isLeftDisjunctive_def]
	THEN REPEAT STRIP_TAC
	THEN XRULE_ASSUM_TAC 
		(simpLib.SIMP_RULE std_ss [] o
	         SPEC (--`\b. if b then (p:'a->bool) else r`--))
	THEN XRULE_ASSUM_TAC 
		(simpLib.SIMP_RULE std_ss [] o
	         SPEC (--`\b. if b then (q:'b->bool) else s`--))
	THEN REWRITE_TAC [lemma, TT_def]
	THEN FIRST_ASSUM MATCH_MP_TAC 
	THEN RW_TAC std_ss []
	end
	) ;
    
val Cancelation_thm = store_thm 
  ("Cancelation_thm",
   --`isTransitive leadsto  /\ isDisjunctive (TT:bool->bool) leadsto /\
         leadsto q q /\ leadsto p (q OR r) /\ leadsto r s
         ==>
         leadsto p (q OR s)`--,
	PROVE_TAC [isTransitive_def, leadsto_SimpleDisjunction_thm]) ;

(*--------------------------------------------------------------------------------

Here is the bounded progres law. It says that if from p a program can
progress to either q or it decreases the value of some well-founded
metric, then eventually q will actually be reached. So, something like
this:

             !m. (p /\ M=m) --> ((p /\ M<m) \/ q)
           ---------------------------------------
                          p --> q

We'll define well-foundedness of a relation in terms of its admitance
to well-founded induction. This notion has been proven to be
equivalent with the standard notion of well-foundedness (well, if
you're not convinced, see also my thesis).

---------------------------------------------------------------------------------*)

val ADMIT_WF_INDUCTION_def 
	= 
	Define
   	`ADMIT_WF_INDUCTION R =
    	 !Property. 
		(!y. (!x. R x y ==> Property x) ==> Property y) 
		==> 
		(!x. Property x)` ;


val MRel    = ty_antiq (Type `:'m->'m->bool`) ;
val Metric  = ty_antiq (Type `:'a->'m`) ;

val lemma1 = prove(--`p OR q = q OR p`--, pred_eq_PROVE_TAC) ;
val lemma2 = prove(--`p AND (\s. LESS (M s) y) 
		= 
		??x::(\x. LESS x y). (\x. p AND (\s. M s = x)) x`--,
		pred_eq_PROVE_TAC) ;	

val BOUNDED_REACH_lemma  
	= 
	prove
  	(--`ADMIT_WF_INDUCTION LESS /\        
	    isTransitive (leadsto:^UOP) /\ 
	    isDisjunctive (TT:bool->bool) leadsto /\
	    (!W. (?i:'m. W i) ==> isDisjunctive W leadsto) /\
            leadsto q q /\
            (!m:'m. leadsto (p AND (\s. M s = m)) 
				 ((p AND (\s. LESS (M s) m)) OR q))
	    ==>
	    (!m. leadsto (p AND (\s. M s = m)) q)`--,

    STRIP_TAC
    THEN SUBGOAL_THEN
	 (--`ADMIT_WF_INDUCTION (LESS:^MRel)`--)
	 (ASSUME_TAC o 
	  simpLib.SIMP_RULE std_ss [] o
	  SPEC (--`\m:'m. (leadsto:^UOP) (p AND (\s. M s = m)) q`--) o
	  REWRITE_RULE [ADMIT_WF_INDUCTION_def])
    THENL [RW_TAC std_ss [], ALL_TAC ]
    THEN FIRST_ASSUM MATCH_MP_TAC
    THEN REPEAT STRIP_TAC
    THEN ASM_CASES_TAC (--`?x. (LESS:^MRel) x y`--)
    THENL
	[ ALL_TAC
	 ,(* case 2: y is a bottom element *)
         XRULE_ASSUM_TAC (SPEC (--`y:'m`--))
	 THEN SUBGOAL_THEN (--`(\s:'a. (LESS:^MRel) (M s) y) = FF`--)
	      (fn th => XRULE_ASSUM_TAC (REWRITE_RULE [th]))
	 THENL
	    [ valid_PROVE_TAC,
	      PROVE_TAC [prove(--`p AND FF OR q = q`--, pred_eq_PROVE_TAC)]
	    ]
	]
   (* continue with case 1: y is not a bottom element *)
   THEN SUBGOAL_THEN 
	 (--`(leadsto:^UOP)
		(p AND (\s. (LESS:^MRel) (M s) (y:'m)))
		 q`--)
	  ASSUME_TAC 
    THENL
	[ ALL_TAC,
         (* with the lemma the original goal can be proven: *)
	 SUBST1_TAC (prove(--`(q:'a->bool) = q OR q`--, pred_eq_PROVE_TAC))
	 THEN MATCH_MP_TAC (GEN_ALL Cancelation_thm)
	 THEN EXISTS_TAC (--`p AND (\s:'a. (LESS:^MRel) (M s) (y:'m))`--)
	 THEN ONCE_REWRITE_TAC [lemma1]
	 THEN RW_TAC std_ss []
	]
    (* continue with proving the lemma *)
    THEN ONCE_REWRITE_TAC [lemma2]
    THEN SUBGOAL_THEN (--`isDisjunctive (\x. (LESS:^MRel) x y) (leadsto:^UOP)`--)
	 ASSUME_TAC
    THENL [RW_TAC std_ss [], ALL_TAC ]
    THEN SUBGOAL_THEN 
         (--`isLeftDisjunctive (\x. (LESS:^MRel) x y) (leadsto:^UOP)`--)
 	 ASSUME_TAC
    THENL [RW_TAC std_ss [LeftDisjuction_lemma], ALL_TAC ]
    THEN FIRST_ASSUM
	 (MATCH_MP_TAC o REWRITE_RULE [isLeftDisjunctive_def])
    THEN RW_TAC std_ss []
    ) ;

val lemma = prove
   (Term `p = (??m:'m::TT. (\m. p AND (\s:'a. M s = m)) m)`,
	pred_eq_PROVE_TAC) ;


val BOUNDED_REACH_thm = store_thm 
  ("BOUNDED_REACH_thm",
   --`ADMIT_WF_INDUCTION (LESS:^MRel) /\        
	isTransitive (leadsto:^UOP) /\ 
	isDisjunctive (TT:bool->bool) leadsto /\
	(!W. (?i:'m. W i) ==> isDisjunctive W leadsto) /\
	leadsto q q /\
	(!m:'m. leadsto (p AND (\s. M s = m)) 
		        ((p AND (\s. LESS (M s) m)) OR q))
	==>
        leadsto p q`--,
   STRIP_TAC
   THEN SUBST1_TAC lemma
   THEN SUBGOAL_THEN 
	(--`isLeftDisjunctive (TT:'m->bool) (leadsto:^UOP)`--)
        (ASSUME_TAC o REWRITE_RULE [isLeftDisjunctive_def])
   THENL [RW_TAC std_ss [LeftDisjuction_lemma, TT_def], ALL_TAC]
   THEN FIRST_ASSUM MATCH_MP_TAC
   THEN RW_TAC std_ss [TT_def]
   THEN MATCH_MP_TAC (GEN_ALL BOUNDED_REACH_lemma)
   THEN EXISTS_TAC (--`LESS:^MRel`--)
   THEN RW_TAC std_ss []
   ) ;

(*--------------------------------------------------------------------------------
    A special case of the reachability by bounded decrement principle is the
    following: 

             !m. (~p /\ M=m) --> M<m \/ p
           --------------------------------
                    true --> p

---------------------------------------------------------------------------------*)

val BOUNDED_ALWAYS_REACH_thm 
	= 
	store_thm ("BOUNDED_ALWAYS_REACH_thm",
	--`ADMIT_WF_INDUCTION (LESS:^MRel) /\        
	isTransitive (leadsto:^UOP) /\ 
	isDisjunctive (TT:bool->bool) leadsto /\
	(!W. (?i:'m. W i) ==> isDisjunctive W leadsto) /\
	leadsto p p /\
	(!m:'m. leadsto (NOT p AND (\s. M s = m)) 
		        ((\s. LESS (M s) m) OR p))
	==>
        leadsto TT p`--,      
	STRIP_TAC
   	THEN (SUBST1_TAC o prove) (--`(p:'a->bool) = p OR p`--, pred_eq_PROVE_TAC)
    	THEN (SUBST1_TAC o prove) (--`TT=((p:'a->bool) OR (NOT p))`--, pred_eq_PROVE_TAC)
    	THEN MATCH_MP_TAC leadsto_SimpleDisjunction_thm
	THEN RW_TAC std_ss []
	THEN MATCH_MP_TAC (GEN_ALL BOUNDED_REACH_thm)
    	THEN EXISTS_TAC (Term `M:^Metric`)
    	THEN EXISTS_TAC (Term `LESS:^MRel`)
	THEN RW_TAC std_ss [prove(--`NOT p AND q OR p = q OR p`--,
                 	    pred_eq_PROVE_TAC)]
	) ;



(*--------------------------------------------------------------------------------

The previously proven laws for leadsto assumes only transitivity and
disjunctivity. Laws like Progress Safety Progress (PSP) and Completion
requires the minimality property of leadsto. That is, it has to be the
least transitive and disjunctive closure of some ensures-like
relation.

The relation GLEADSTO (general leadsto) is defined to be such a
closure. Notice that this is a very general progress relation, since
we don't specify what the underlying ensures relation.

We give two definitions for GLEADSTO ... both, as proven later, are
equivalent.

---------------------------------------------------------------------------------*)

val GLEADSTO_def = Define
	`GLEADSTO ensures p q 
	 = 
	 !U. isSubRelationOf ensures U /\ 
	     isTransitive U /\ 
	     isLeftDisj U
	     ==> 
	     U p q` ;

val GLEADSTO2_def = Define
	`GLEADSTO2 ensures p q 
	 = 
	 !U. isSubRelationOf ensures U /\ 
	     isTransitive2 ensures U /\ 
	     isLeftDisj U
	     ==> 
	     U p q` ;

(*-----------------------------------------------------------------------------------

So, GLEADSTO satisfies the obvious properties:
	
	*It includes the underlying ensures.
	*It includes implication
	*It is transitive.
	*It is disjunctive.

-----------------------------------------------------------------------------------*)

val ensures_to_GLEADSTO_lift_thm
	=
	store_thm
	("ensures_to_GLEADSTO_lift_thm",
	--`ensures p q
	   ==>
	   GLEADSTO ensures p q`--,
	RW_TAC std_ss [isSubRelationOf_def, GLEADSTO_def]
	) ;

val implies_to_GLEADSTO_lift_thm
	=
	store_thm
	("implies_to_GLEADSTO_lift_thm",
	--`isSubRelationOf implies ensures /\
	   implies p q
	   ==>
	   GLEADSTO ensures p q`--,
	RW_TAC std_ss [isSubRelationOf_def, GLEADSTO_def]
	) ;

val IMP_to_GLEADSTO_lift_thm
	=
	store_thm
	("IMP_to_GLEADSTO_lift_thm",
	--`includesIMP implies /\
	   isSubRelationOf implies ensures /\
	   inDomain implies p /\ inDomain implies q /\
	   valid (p IMP q)
	   ==>
	   GLEADSTO ensures p q`--,
	RW_TAC std_ss [isSubRelationOf_def, includesIMP_def, GLEADSTO_def]
	) ;

val GLEADSTO_transitive_thm
	=
	store_thm
	("GLEADSTO_transitive_thm",
	--`isTransitive (GLEADSTO ensures)`--,
	RW_TAC std_ss [isTransitive_def, GLEADSTO_def]
	THEN RES_TAC THEN RES_TAC
	) ;

val GLEADSTO_transitive_thm_morph1
	=
	store_thm
	("GLEADSTO_transitive_thm_morph1",
	--`GLEADSTO ensures p q /\
	   GLEADSTO ensures q r
	   ==>
	   GLEADSTO ensures p r`--,
	PROVE_TAC [REWRITE_RULE [isTransitive_def] GLEADSTO_transitive_thm]) ;


val GLEADSTO_LeftDisj_thm
	=
	store_thm
	("GLEADSTO_LeftDisj_thm",
	--`isLeftDisj (GLEADSTO ensures)`--,
	RW_TAC std_ss [isLeftDisj_def, GLEADSTO_def]
	THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC
	THEN PROVE_TAC []) ;

val GLEADSTO_LeftDisj_thm_morph1
	=
	store_thm
	("GLEADSTO_LeftDisj_thm_morph1",
	--`W i /\
	   (!p. W p ==> GLEADSTO ensures p q)
	   ==> 
	   GLEADSTO ensures (??p::W. p) q`--,
	STRIP_TAC
	THEN IMP_RES_TAC (REWRITE_RULE [isLeftDisj_def] 
		   	  GLEADSTO_LeftDisj_thm)
	) ;


val GLEADSTO_LeftDisj_thm_morph2
	=
	store_thm
	("GLEADSTO_LeftDisj_thm_morph2",
	--`W i /\
	   (!i:'index. W i ==> GLEADSTO (ensures:^UOP) (f i) q)
	   ==> 
	   GLEADSTO ensures (??i::W. f i) q`--,
	STRIP_TAC
	THEN (ASSUME_TAC o SPEC(--`ensures:^UOP`--) o GEN_ALL)
	      GLEADSTO_LeftDisj_thm
	THEN IMP_RES_TAC (REWRITE_RULE [isLeftDisjunctive_def] 
			  LeftDisj_imp_LeftDisjunction_thm)
	) ;

val GLEADSTO_LeftDisj_thm_morph3
	=
	store_thm
	("GLEADSTO_LeftDisj_thm_morph3",
	--`isLeftDisjunctive (TT:bool->bool) (GLEADSTO ensures)`--,

	PROVE_TAC [GLEADSTO_LeftDisj_thm,
		   LeftDisj_imp_LeftDisjunction_thm,
		   prove(--`TT T`--, valid_PROVE_TAC)]
	) ;


(*-------------------------------------------------------------------------------

Being the least closure, GLEADTO induces an induction principle
(called leadsto induction in CM88).
      
-------------------------------------------------------------------------------*)

val GLEADSTO_induction1_thm = store_thm 
	("GLEADSTO_induction1_thm",
	--`isSubRelationOf ensures U /\
	   isTransitive U /\
	   isLeftDisj U
	   ==>
	   isSubRelationOf (GLEADSTO ensures) U`--,

	   RW_TAC std_ss [isSubRelationOf_def, GLEADSTO_def]) ;


val GLEADSTO_induction2_thm = store_thm 
	("GLEADSTO_induction2_thm",
	--`isSubRelationOf ensures (rINTER (GLEADSTO ensures) U) /\
	   isTransitive (rINTER (GLEADSTO ensures) U) /\
	   isLeftDisj   (rINTER (GLEADSTO ensures) U)
	   ==>
	   isSubRelationOf (GLEADSTO (ensures:^UOP)) U`--,

	STRIP_TAC
	THEN SUBGOAL_THEN
		(--`isSubRelationOf 
			(GLEADSTO (ensures:^UOP))
			(rINTER   (GLEADSTO ensures) U)`--)
		ASSUME_TAC
	THENL [ ALL_TAC,
		UNDISCH_ALL_TAC
		THEN RW_TAC std_ss [isSubRelationOf_def, rINTER_def] ]
	THEN PROVE_TAC [GLEADSTO_induction1_thm]
	) ;

val GLEADSTO2_induction1_thm = store_thm 
	("GLEADSTO2_induction1_thm",
	--`isSubRelationOf ensures U /\
	   isTransitive2 ensures U /\
	   isLeftDisj U
	   ==>
	   isSubRelationOf (GLEADSTO2 ensures) U`--,

	   RW_TAC std_ss [isSubRelationOf_def, GLEADSTO2_def]) ;

(*-------------------------------------------------------------------------------

The following theorems states the condition on which GLEADSTO will have
a proper domain.

-------------------------------------------------------------------------------*)


val GLEADSTO_hasProperDomain_lemma
	=
	store_thm
	("GLEADSTO_hasProperDomain_lemma",
	--`isClosed_under_PredOPS (inDomain implies) /\
	    hasProperDomain (ensures:^UOP) /\
	   (!p. inDomain ensures p ==> inDomain implies p)
	   ==>
	   (!p q. GLEADSTO ensures p q ==>  inDomain implies p /\ inDomain implies q)
	`--,

	STRIP_TAC
	THEN SUBGOAL_THEN
		(--`isSubRelationOf (GLEADSTO ensures)
			(\p q. inDomain implies p /\
			       inDomain implies q)`--)
		(fn th=> PROVE_TAC
				[BETA_RULE 
				 (REWRITE_RULE [isSubRelationOf_def] th)
				]
		)
	THEN MATCH_MP_TAC GLEADSTO_induction1_thm 
	THEN REPEAT CONJ_TAC
	THENL
	[ (** 1 **)
	  RW_TAC std_ss [isSubRelationOf_def]
	  THEN PROVE_TAC [hasProperDomain_def],
	  (** 2 **)
	  RW_TAC std_ss [isTransitive_def],
	  (** 3 **)
	  RW_TAC std_ss [isLeftDisj_def]
	  THEN PROVE_TAC [domain_closed_qOR_lemma2]
	]
	) ;

val GLEADSTO_hasProperDomain_thm 
	=
	store_thm
	("GLEADSTO_hasProperDomain_thm",
	--`includesIMP implies /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   isSubRelationOf implies ensures /\
	   hasProperDomain (ensures:^UOP) /\
	   (!p. inDomain ensures p ==>  inDomain implies p)
	   ==> 
	   hasProperDomain (GLEADSTO ensures)`--,

	STRIP_TAC
	THEN REWRITE_TAC [hasProperDomain_def]
	THEN SUBGOAL_THEN
		(--`isSubRelationOf (GLEADSTO ensures)
			(\p q. inDomain (GLEADSTO ensures) p /\
			       inDomain (GLEADSTO ensures) q)`--)
		(fn th=> PROVE_TAC
				[BETA_RULE 
				 (REWRITE_RULE [isSubRelationOf_def] th)
				]
		)
	THEN MATCH_MP_TAC GLEADSTO_induction1_thm 
	THEN REPEAT CONJ_TAC
	THENL
	[ (** (1) ens lifting **)
	  RW_TAC std_ss [isSubRelationOf_def]
	  THEN PROVE_TAC [hasProperDomain_def, 
			  inDomain_def, 
			  ensures_to_GLEADSTO_lift_thm],
	  (** (2) transitivity **)
	  RW_TAC std_ss [isTransitive_def],
	  (** (3) left disjunctivity **)
	  RW_TAC std_ss [isLeftDisj_def, inDomain_def]
	  THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_LeftDisj_thm_morph1)
	  THEN EXISTS_TAC (--`p:^Pred`--)
	  THEN RW_TAC std_ss []
	  THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_transitive_thm_morph1)
	  THEN EXISTS_TAC (--`p':^Pred`--)
	  THEN RW_TAC std_ss []	  
	  THEN MATCH_MP_TAC IMP_to_GLEADSTO_lift_thm
	  THEN RW_TAC std_ss [inDomain_def]
	  THENL
	   [ PROVE_TAC [GLEADSTO_hasProperDomain_lemma, inDomain_def] ,
	     XRULE_ASSUM_TAC (REWRITE_RULE [includesIMP_def])
	     THEN FIRST_ASSUM MATCH_MP_TAC
	     THEN RW_TAC std_ss []
	     THEN TRY (PROVE_TAC [GLEADSTO_hasProperDomain_lemma, 
				  domain_closed_qOR_lemma2])
	     THEN valid_PROVE_TAC,
	     valid_PROVE_TAC
	   ]
	]
	) ;

(*-------------------------------------------------------------------------------

Derived properties of GLEADSTO:

	*substitution
	*general disjunctivity
	*calcelation
	*bounded progress laws
    
-------------------------------------------------------------------------------*)

val GLEADSTO_Subst_thm 
	=
	store_thm
	("GLEADSTO_Subst_thm",
	--`isSubRelationOf implies ensures /\
	   implies a p /\ implies q b /\
	   GLEADSTO ensures p q
	   ==>
	   GLEADSTO ensures a b`--,

	PROVE_TAC [leadsto_Subst_thm, isSubRelationOf_def,
		   GLEADSTO_transitive_thm,
		   ensures_to_GLEADSTO_lift_thm] 
	) ;


val GLEADSTO_Disj_thm
	=
	store_thm
	("GLEADSTO_Disj_thm",
	--`includesIMP implies /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   isSubRelationOf implies ensures /\
	   hasProperDomain (ensures:^UOP) /\
	   (!p. inDomain ensures p ==> inDomain implies p) /\
	   W (i:'index) 
	   ==>
	   isDisjunctive W (GLEADSTO ensures)`--,

	STRIP_TAC
	THEN MATCH_MP_TAC LeftDisj_imp_Disjunction_thm
	THEN RW_TAC std_ss [GLEADSTO_LeftDisj_thm, GLEADSTO_transitive_thm]
	THENL
	[ PROVE_TAC [isSubRelationOf_def,ensures_to_GLEADSTO_lift_thm],
	  PROVE_TAC [inDomain_def, 
		     hasProperDomain_def, 
		     GLEADSTO_hasProperDomain_lemma],
	  PROVE_TAC [GLEADSTO_hasProperDomain_thm]
	]
	) ;

val GLEADSTO_RightDisj_thm
	=
	store_thm 
  	("GLEADSTO_RightDisj_thm",
	--`includesIMP implies /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   isSubRelationOf implies ensures /\
	   hasProperDomain (ensures:^UOP) /\
	   (!p. inDomain ensures p ==> inDomain implies p) /\
	   W i /\
	   (!i. W i ==> GLEADSTO ensures p (f i))
           ==>
           GLEADSTO ensures p (??i::W. f i)`--,

	STRIP_TAC
	THEN MATCH_MP_TAC leadsto_GeneralRightDisjunction_thm
	THEN PROVE_TAC [GLEADSTO_Disj_thm]
	) ;


val GLEADSTO_SimpleDisjunction_thm = store_thm 
  ("GLEADSTO_SimpleDisjunction_thm",
    --`includesIMP implies /\
        isClosed_under_PredOPS (inDomain implies) /\
	isSubRelationOf implies ensures /\
	hasProperDomain (ensures:^UOP) /\
	(!p. inDomain ensures p ==> inDomain implies p) /\
	GLEADSTO ensures p q /\ GLEADSTO ensures r s
       ==>
       GLEADSTO ensures (p OR r) (q OR s)`--,

	STRIP_TAC
	THEN MATCH_MP_TAC leadsto_SimpleDisjunction_thm
	THEN PROVE_TAC [GLEADSTO_Disj_thm,
			prove (--`TT T`--, valid_PROVE_TAC) ]
	) ;


val GLEADSTO_Cancelation_thm = store_thm 
  ("GLEADSTO_Cancelation_thm",
   --`includesIMP implies /\
        isClosed_under_PredOPS (inDomain implies) /\
	isSubRelationOf implies ensures /\
	hasProperDomain (ensures:^UOP) /\
	(!p. inDomain ensures p ==> inDomain implies p) /\
        inDomain (GLEADSTO ensures) q /\
	GLEADSTO ensures p (q OR r) /\ 
	GLEADSTO ensures r s
        ==>
        GLEADSTO ensures p (q OR s)`--,

	STRIP_TAC
	THEN MATCH_MP_TAC Cancelation_thm 
	THEN RW_TAC std_ss [GLEADSTO_LeftDisj_thm, GLEADSTO_transitive_thm]
	THENL
	[ MATCH_MP_TAC GLEADSTO_Disj_thm
	  THEN PROVE_TAC [prove (--`TT i`--, valid_PROVE_TAC) ],
	  PROVE_TAC [inDomain_def]
	]
	) ;

val GLEADSTO_BOUNDED_REACH_thm = store_thm 
  ("GLEADSTO_BOUNDED_REACH_thm",
   --`includesIMP implies /\
        isClosed_under_PredOPS (inDomain implies) /\
	isSubRelationOf implies ensures /\
	hasProperDomain (ensures:^UOP) /\
	(!p. inDomain ensures p ==> inDomain implies p) /\
	ADMIT_WF_INDUCTION (LESS:^MRel) /\        
	inDomain (GLEADSTO ensures) q /\
	(!m:'m. GLEADSTO ensures (p AND (\s. M s = m)) 
		        	 ((p AND (\s. LESS (M s) m)) OR q))
	==>
        GLEADSTO ensures p q`--,

	STRIP_TAC
	THEN MATCH_MP_TAC BOUNDED_REACH_thm
	THEN RW_TAC std_ss [GLEADSTO_LeftDisj_thm, GLEADSTO_transitive_thm]
	THENL
	[ MATCH_MP_TAC GLEADSTO_Disj_thm
	  THEN PROVE_TAC [prove (--`TT i`--, valid_PROVE_TAC) ],
	  PROVE_TAC [GLEADSTO_Disj_thm],
	  PROVE_TAC [inDomain_def]
	]
	) ;

val GLEADSTO_BOUNDED_ALWAYS_REACH_thm 
	= 
	store_thm ("GLEADSTO_BOUNDED_ALWAYS_REACH_thm",
	--`includesIMP implies /\
        isClosed_under_PredOPS (inDomain implies) /\
	isSubRelationOf implies ensures /\
	hasProperDomain (ensures:^UOP) /\
	(!p. inDomain ensures p ==> inDomain implies p) /\
	ADMIT_WF_INDUCTION (LESS:^MRel) /\        
	inDomain (GLEADSTO ensures) p /\
	(!m:'m. GLEADSTO ensures (NOT p AND (\s. M s = m)) 
		        	 ((\s. LESS (M s) m) OR p))
	==>
        GLEADSTO ensures TT p`--,    

	STRIP_TAC
	THEN MATCH_MP_TAC BOUNDED_ALWAYS_REACH_thm
	THEN RW_TAC std_ss [GLEADSTO_LeftDisj_thm, GLEADSTO_transitive_thm]
	THENL
	[ MATCH_MP_TAC GLEADSTO_Disj_thm
	  THEN PROVE_TAC [prove (--`TT i`--, valid_PROVE_TAC) ],
	  PROVE_TAC [GLEADSTO_Disj_thm],
	  PROVE_TAC [inDomain_def]
	]
	) ;


(*-------------------------------------------------------------------------------

So, here is the theorem stating that both definitions for GLEADSTO are
equivalent.

-------------------------------------------------------------------------------*)
 
val GLEADSTO2_imp_GLEADSTO_lemma
	=
	prove(
	--`isSubRelationOf (GLEADSTO2 ensures)
			   (GLEADSTO (ensures:^UOP))`--,

	MATCH_MP_TAC (GEN_ALL GLEADSTO2_induction1_thm)
	THEN RW_TAC std_ss
		[isSubRelationOf_def,
		 ensures_to_GLEADSTO_lift_thm,
		 GLEADSTO_LeftDisj_thm, 
		 isTransitive2_def ]
	THEN PROVE_TAC 
		[ensures_to_GLEADSTO_lift_thm,
		 REWRITE_RULE [isTransitive_def] GLEADSTO_transitive_thm]
	) ;


val lemma = prove
  (Term  `(X x y /\ X y z ==> X x z) =
          (X x y ==> ((\x y. X y z ==> X x z) x y))`,
   PROVE_TAC[]) ;


val GLEADSTO_imp_GLEADSTO2_lemma
	=
	prove(
	--`isSubRelationOf (GLEADSTO  ensures)
			   (GLEADSTO2 (ensures:^UOP))`--,

	MATCH_MP_TAC (GEN_ALL GLEADSTO_induction1_thm)
	THEN REPEAT STRIP_TAC
	THENL
	[ (* 1: prove ensures lifting for GLEADSTO2 *)
	  RW_TAC std_ss [isSubRelationOf_def, GLEADSTO2_def],
	
	  (* 2: prove GLEADSTO2 is generally transitive *)
	  REWRITE_TAC [isTransitive_def, lemma]
	  THEN REPEAT GEN_TAC
	  THEN SPEC_TAC (--`q:'a->bool`--,--`q:'a->bool`--)
	  THEN SPEC_TAC (--`p:'a->bool`--,--`p:'a->bool`--)
	  THEN REWRITE_TAC [SYM (SPEC_ALL isSubRelationOf_def)]
	  THEN MATCH_MP_TAC (GEN_ALL GLEADSTO2_induction1_thm)
	  THEN REPEAT CONJ_TAC
	  THENL
	  [ (* 2.1 *)
	     RW_TAC std_ss [isSubRelationOf_def,
			    isTransitive2_def,
			    GLEADSTO2_def ]
	     THEN RES_TAC THEN RES_TAC,
	     (* 2.2 *)
	     RW_TAC std_ss [isSubRelationOf_def,
			    isTransitive2_def,
			    GLEADSTO2_def ]
	     THEN RES_TAC THEN RES_TAC,
	     (* 2.3 *)
	     REWRITE_TAC [isLeftDisj_def,
			    GLEADSTO2_def ]
	     THEN BETA_TAC THEN REPEAT STRIP_TAC
	     THEN RES_TAC
	     THEN FIRST_ASSUM MATCH_MP_TAC
	     THEN RW_TAC std_ss []
	  ],
	  (* 3: prove GLEADST2 is left disjunctive *)
	  RW_TAC std_ss [isLeftDisj_def, GLEADSTO2_def]
	  THEN RES_TAC THEN FIRST_ASSUM MATCH_MP_TAC
	  THEN PROVE_TAC []
	]
	) ;


val GLEADSTO_equal_GLEADSTO2_thm =
	store_thm
	("GLEADSTO_equal_GLEADSTO2_thm",
	--`GLEADSTO (ensures:^UOP)
	   =
	   GLEADSTO2 ensures`--,
	CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
	THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
	THEN PROVE_TAC 
		[REWRITE_RULE [isSubRelationOf_def] GLEADSTO2_imp_GLEADSTO_lemma,
		 REWRITE_RULE [isSubRelationOf_def] GLEADSTO_imp_GLEADSTO2_lemma
		]
	) ;

(*-----------------------------------------------------------------------------

Okay... here are then the 'heavy weight' laws for leadsto-like
properties:

	*Monotonicity: if ensures1 is included in ensures2, then the
	leadsto closure of ensures1 will also be included in the
	leadsto closure of ensures 2.

	*the PSP law.
	
	*the completion law ... (most difficult one to prove!)

-----------------------------------------------------------------------------*)


val isSubRelationOf_transitive_lemma
	=
	prove
	(--`isSubRelationOf U V /\ isSubRelationOf V W
	    ==>
	    isSubRelationOf U W`--,
	 RW_TAC std_ss [isSubRelationOf_def]) ;


val GLEADSTO_inclusion_thm
	=
	store_thm
	("GLEADSTO_inclusion_thm",
	--`isSubRelationOf ensures1 (ensures2:^UOP)
	   ==>
	   isSubRelationOf (GLEADSTO ensures1)
			   (GLEADSTO ensures2)`--,

	STRIP_TAC
	THEN MATCH_MP_TAC (GLEADSTO_induction1_thm)
	THEN RW_TAC std_ss [GLEADSTO_transitive_thm,
			    GLEADSTO_LeftDisj_thm]
	THEN UNDISCH_ALL_TAC
	THEN RW_TAC std_ss [isSubRelationOf_def]
	THEN PROVE_TAC [ensures_to_GLEADSTO_lift_thm]
	) ;


val satisfiesPSP_def
	=
	Define
	`satisfiesPSP progress unless
	 =
	 !p q a b. progress p q /\ unless a b 
		   ==> 
		   progress (p AND a) (q AND a OR b)` ;


val lemma = prove
	(--`(??p::W. p) AND (a:'a->bool) 
	     = 
	     (??p::(\y. (?q. W q /\ (y = q AND a))). p)`--,
	pred_eq_decompose_TAC 
	THEN CONV_TAC (DEPTH_CONV FUN_EQ_CONV)
	THEN STRIP_TAC THEN EQ_TAC
	THEN RW_TAC std_ss [] THEN TRY (PROVE_TAC [])
	THEN EXISTS_TAC (--`(\x:'a. i x /\ a x)`--)
	THEN BETA_TAC 
	THEN PROVE_TAC []) ;

val PSP_thm
	=
	store_thm
	("PSP_thm",
	--`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies ensures /\
	   isDisjunctive (TT:bool->bool) (GLEADSTO ensures) /\
	   satisfiesPSP (ensures:^UOP) unless
	   ==>
	   satisfiesPSP (GLEADSTO ensures) unless`--,
	REWRITE_TAC [satisfiesPSP_def] 
	THEN STRIP_TAC
	THEN SUBGOAL_THEN (--`!a b. (unless:^UOP) a b ==> ensures b b`--)
	     ASSUME_TAC
	THENL [ PROVE_TAC 
			[includesIMP_def, 
			 isSubRelationOf_def, 
			 hasProperDomain_def, 
			 inDomain_def,
			 prove(--`valid (b IMP b)`--, valid_PROVE_TAC)
		 	],
		ALL_TAC ]
	THEN SUBGOAL_THEN
	     (--`!p q.
	            GLEADSTO (ensures:^UOP) p q 
		    ==>
		    (\p q. (!a b. unless a b 
		                  ==>
		                  GLEADSTO ensures (p AND a) (q AND a OR b))
	            ) p q
	      `--)
	    (fn th=> PROVE_TAC [BETA_RULE th])
	THEN REWRITE_TAC [SYM (SPEC_ALL isSubRelationOf_def)]
	THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_induction1_thm)
	THEN REPEAT CONJ_TAC
	THENL
	[ (* 1: ensures lifting case *)
	  RW_TAC std_ss [isSubRelationOf_def, ensures_to_GLEADSTO_lift_thm],
	  (* 2: transitivity case *)
	  RW_TAC std_ss [isTransitive_def]
	  THEN RES_TAC
	  THEN SUBGOAL_THEN
	       (--`GLEADSTO (ensures:^UOP) 
			    (q AND a OR b) 
			    (r AND a OR b)`--)
	       ASSUME_TAC
	  THENL [ ALL_TAC, PROVE_TAC [GLEADSTO_transitive_thm_morph1]]
	  THEN SUBST1_TAC 
		(prove(--`r AND a OR b = (r AND a OR b) OR (b:'a->bool)`--,
			  valid_PROVE_TAC))
	  THEN MATCH_MP_TAC (GEN_ALL leadsto_SimpleDisjunction_thm)
	  THEN RW_TAC std_ss [ensures_to_GLEADSTO_lift_thm], 
	   (* 3: disjunctivity case *)
	  REWRITE_TAC [isLeftDisj_def]
	  THEN BETA_TAC THEN REPEAT STRIP_TAC
	  THEN REWRITE_TAC [lemma]
	  THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_LeftDisj_thm_morph1)
	  THEN RW_TAC std_ss []
	  THEN PROVE_TAC []
	]
	) ;


val PSP_thm_alt =
	CONV_RULE (RAND_CONV (REWR_CONV satisfiesPSP_def)
		     THENC
		     NORMALIZE_FOR_MP_CONV)
	 PSP_thm ;


val PSP_thm_x = store_thm
	("PSP_thm_x",
	--`includesIMP implies /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   hasProperDomain ensures /\
	   isSubRelationOf implies ensures /\
	   (!p. inDomain ensures p ==> inDomain implies p) /\
	   satisfiesPSP (ensures:^UOP) unless /\
	   GLEADSTO ensures p q /\
           unless a b
	   ==>
	   GLEADSTO ensures (p AND a) ((q AND a) OR b)`--,

	STRIP_TAC
	THEN MATCH_MP_TAC PSP_thm_alt
	THEN RW_TAC std_ss []
	THEN PROVE_TAC [prove(--`TT T`--, valid_PROVE_TAC),
			GLEADSTO_Disj_thm]
	) ;


(*** Now the completion law ... ***)

(*-----------------------------------------------------------------------------

Completion:

	p |--> q     and   q unless b

has actually a quite abstract (intuitive) interpretation. It's a
generalization of convergence (Pra95).

We will therefore define a derived operator to capture this
abstraction:

	COMPLETES p q b

defined as follows:

-----------------------------------------------------------------------------*)


val COMPLETES_def = Define
	`COMPLETES progress unless p q b
	 =
	 progress p q /\ unless q b` ; 



val satisfiesCompLemma_def = Define
	`satisfiesCompLemma unless ensures leadsto
	 =
	 !p q r s b. 
		ensures p q /\
	 	leadsto r s /\
		unless  s b
		==>
		?mid. leadsto (p AND r) (q AND mid OR
				         q AND s OR b)
		      /\
		      leadsto mid s` 
	;

(*
val satisfiesCompletion_def = Define
	`satisfiesCompletion unless ensures leadsto
	 =
	 !p1 q1 p2 q2 b. 
		COMPLETES ensures unless p1 q1 b /\
	 	COMPLETES leadsto unless p2 q2 b
		==>
		COMPLETES leadsto unless
			  (p1 AND p2) 
			  (q1 AND q2 OR b) b` ;
*)

(*** mesonLib.max_depth := 40 ; ***)

val lemma1 = prove(--`(b:^Pred) = (b OR b) OR b`--, pred_eq_PROVE_TAC) ;


val Completion_preparation
	=
	prove(
	--`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies unless /\
	   isSubRelationOf implies ensures /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   satisfiesUNLESS_Disj unless /\
	   satisfiesPSP ensures unless /\
	   hasProperDomain (ensures:^UOP) /\  (* *)
	   (!p. inDomain ensures p ==> inDomain implies p) /\ (* *)
	   satisfiesCompLemma unless ensures (GLEADSTO ensures)
	   ==>
	   isSubRelationOf
		(GLEADSTO (ensures:^UOP))
		(\p q. GLEADSTO ensures p q /\
		       (unless q b ==>
		       (!r s. COMPLETES (GLEADSTO ensures) unless r s b
			      ==>
			      COMPLETES (GLEADSTO ensures) unless 
					 (p AND r) (q AND s OR  b)
					  b
			))
		)`--,

	STRIP_TAC
	THEN SUBGOAL_THEN
		(--`isDisjunctive (TT:bool->bool) (GLEADSTO (ensures:^UOP))`--)
		ASSUME_TAC
	THENL [PROVE_TAC [prove(--`TT T`--, valid_PROVE_TAC),
			  GLEADSTO_Disj_thm], 
	       ALL_TAC]
	THEN REWRITE_TAC [GLEADSTO_equal_GLEADSTO2_thm]
	THEN MATCH_MP_TAC (GEN_ALL GLEADSTO2_induction1_thm)
	THEN REWRITE_TAC [SYM GLEADSTO_equal_GLEADSTO2_thm]
	THEN REPEAT CONJ_TAC	
	THENL
	[ (*1 : ensures lifting *)

	  RW_TAC std_ss [isSubRelationOf_def, COMPLETES_def]
	  THENL [ (* leadsto part *)
		  PROVE_TAC [ensures_to_GLEADSTO_lift_thm],
		  ALL_TAC,
		  ALL_TAC
		]
	  THENL
	  [ (* 1.1 progress part of the completion *)
	    SUBST1_TAC 
		 (prove(--`y AND s OR (b:^Pred)
			   =
		           (y AND s OR b) OR (y AND s OR b)`--,
			valid_PROVE_TAC))
	    THEN MATCH_MP_TAC (GEN_ALL Cancelation_thm)
	    THEN RW_TAC std_ss [GLEADSTO_transitive_thm]
	    THEN XRULE_ASSUM_TAC (REWRITE_RULE [satisfiesCompLemma_def])
	    THEN RES_TAC
	    THEN EXISTS_TAC (--`y AND (mid:^Pred)`--)
	    THEN REPEAT CONJ_TAC
	    THENL
	    [ (* 1.1.1 *)
	       MATCH_MP_TAC IMP_to_GLEADSTO_lift_thm
	       THEN RW_TAC std_ss []
	       THENL [ALL_TAC, ALL_TAC, valid_PROVE_TAC]
	       THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
	       THEN PROVE_TAC domain_closedness_lemmas,
	       (* 1.1.2 *)
	       PROVE_TAC [prove(--`p OR q OR r = r OR p OR q`--, 
		                   pred_eq_PROVE_TAC)],
	       (* 1.1.3 *)
                ONCE_REWRITE_TAC [prove(--`p AND q =  q AND p`--, 
		                   pred_eq_PROVE_TAC)]
	        THEN (MATCH_MP_TAC o 
			CONV_RULE NORMALIZE_FOR_MP_CONV o
			REWRITE_RULE [satisfiesPSP_def])
			PSP_thm
		THEN RW_TAC std_ss []
		THEN PROVE_TAC [satisfiesPSP_def]
	    ],

	    (* 1.2 unless part *)
	    SUBGOAL_THEN (--`(unless:^UOP) b b`--) ASSUME_TAC
	    THENL
	      [ PROVE_TAC [prove(--`valid (b IMP b)`--, valid_PROVE_TAC),
		           hasProperDomain_def,
		      	   isSubRelationOf_def,
	      		   includesIMP_def],
	        ALL_TAC ]
	    THEN SUBST1_TAC lemma1
	    THEN MATCH_MP_TAC unless_simple_disjunction_thm
	    THEN ASM_REWRITE_TAC [SYM lemma1]
	    THEN PROVE_TAC [unless_simple_conjunction_thm]
	  ],

	  (*2 : transitivity *)

	  RW_TAC std_ss [isTransitive2_def,COMPLETES_def]
	  THENL [(* the leadsto part *)
		 PROVE_TAC [ensures_to_GLEADSTO_lift_thm,
			    GLEADSTO_transitive_thm_morph1],
		 ALL_TAC, 
		 ALL_TAC ]
	  THEN RW_TAC std_ss []
	  THENL
	   [ (* 2.1 progress part *)
	     SUBGOAL_THEN
		(--`?mid.
               		GLEADSTO ensures (p AND r') (q AND mid OR q AND s OR b) /\
               		GLEADSTO (ensures:^UOP) mid s`--)

	        (fn th => ASSUME_TAC th 
			  THEN UNDISCH_TAC (concl th) 
			  THEN STRIP_TAC)

	     THENL [PROVE_TAC [satisfiesCompLemma_def], ALL_TAC]
	     THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_transitive_thm_morph1)
	     THEN EXISTS_TAC
		  (--`q AND mid OR q AND s OR (b:^Pred)`--)
	     THEN RW_TAC std_ss []

	     THEN SUBST1_TAC 
			(prove(--`r AND s OR (b:^Pred) =
  			          (r AND s OR b) OR (r AND s OR b) OR b`--,
			       pred_eq_PROVE_TAC))

	     THEN MATCH_MP_TAC leadsto_SimpleDisjunction_thm
	     THEN RW_TAC std_ss []
	     THENL [ ALL_TAC,
		     PROVE_TAC 
			[IMP_to_GLEADSTO_lift_thm,
			 hasProperDomain_def,
			 prove(--`valid (p IMP p)`--, valid_PROVE_TAC)
			]
		    ]

	     THEN MATCH_MP_TAC leadsto_SimpleDisjunction_thm
	     THEN RW_TAC std_ss []
	     THEN (MATCH_MP_TAC o 
			CONV_RULE NORMALIZE_FOR_MP_CONV o
			REWRITE_RULE [satisfiesPSP_def])
			PSP_thm
	     THEN RW_TAC std_ss []
 	     THEN PROVE_TAC [satisfiesPSP_def]
	     ,

	     (* 2.2 unless part *)
	     SUBGOAL_THEN (--`(unless:^UOP) b b`--) ASSUME_TAC
	    THENL
	      [ PROVE_TAC [prove(--`valid (b IMP b)`--, valid_PROVE_TAC),
		           hasProperDomain_def,
		      	   isSubRelationOf_def,
	      		   includesIMP_def],
	        ALL_TAC ]
	    THEN SUBST1_TAC lemma1
	    THEN MATCH_MP_TAC unless_simple_disjunction_thm
	    THEN ASM_REWRITE_TAC [SYM lemma1]
	    THEN PROVE_TAC [unless_simple_conjunction_thm]
	   ]

	  ,

	  (*3 : disjunctivity *)

	  RW_TAC std_ss [COMPLETES_def, isLeftDisj_def] 
	  THENL [(* the leadsto part *)
		 PROVE_TAC [GLEADSTO_LeftDisj_thm_morph1],
		 ALL_TAC,
		 ALL_TAC]
	  THENL
	  [ (* 3.1 progress part of the completion*)
	    REWRITE_TAC	[
		prove(--`(??p::W. p) AND r
			 =
			 (??p::W. (\i. i AND r) p)`--,
			 pred_eq_PROVE_TAC)
		]
	    THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_LeftDisj_thm_morph2)
	    THEN RW_TAC std_ss [] THEN PROVE_TAC [],
	    (* 3.2 unless part *)
	    PROVE_TAC []
	  ]
	  (* end part 3 *)
	]
	) ;

val Completion_preparation2 =
	(CONV_RULE NORMALIZE_FOR_MP_CONV o
		 BETA_RULE o
		 (CONV_RULE o RAND_CONV o DEPTH_CONV o REWR_CONV) isSubRelationOf_def
	)
	Completion_preparation ;

val Completion_preparation3 =
	let
	val th = SPEC_ALL Completion_preparation2
	val (asml,q) = dest_imp (concl th)
	val (q0,q') = dest_conj q
	in
	prove(mk_imp(asml,q'), 
		MATCH_MP_TAC 
			(GEN_ALL (prove(--`(p==> q /\ r) ==> (p==>r)`--,
			 PROVE_TAC[])))
		THEN EXISTS_TAC q0 
		THEN REWRITE_TAC [th]
		)	
	end ;
	
val Completion_preparation4 =
	(CONV_RULE NORMALIZE_FOR_MP_CONV) 
 	 Completion_preparation3 ;


(*-----------------------------------------------------------------------------

Our version of completion law looks slighly different than the
classical formulation of the law (CM88). It's equivalent, but we think
the formulation is nicer.

-----------------------------------------------------------------------------*)

val Completion_thm
	=
	store_thm
	("Completion_thm",
	--`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies unless /\
	   isSubRelationOf implies ensures /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   satisfiesUNLESS_Disj unless /\
	   satisfiesPSP ensures unless /\
	   hasProperDomain (ensures:^UOP) /\  (* *)
	   (!p. inDomain ensures p ==> inDomain implies p) /\ (* *)
	   satisfiesCompLemma unless ensures (GLEADSTO ensures) /\

	   COMPLETES (GLEADSTO ensures) unless p q b /\
	   COMPLETES (GLEADSTO ensures) unless r s b 
	   ==>
           COMPLETES (GLEADSTO ensures) unless 
		     (p AND r) 
		     (q AND s OR  b)
		      b`--,

	   STRIP_TAC
	   THEN MATCH_MP_TAC Completion_preparation4
	   THEN RW_TAC std_ss []
	   THEN PROVE_TAC [COMPLETES_def]
	) ;	


val Completion_preparation5 =
	let
	val th = SPEC_ALL ((REWRITE_RULE [COMPLETES_def]) Completion_preparation4)
	val (asml,q) = dest_imp (concl th)
	val (q0,q1) = dest_conj q
	in
	prove(mk_imp(asml,q0), 
		MATCH_MP_TAC 
			(GEN_ALL (prove(--`(p==> q /\ r) ==> (p==>q)`--,
			 PROVE_TAC[])))
		THEN EXISTS_TAC q1 
		THEN REWRITE_TAC [th]
		)	
	end ;

(*-----------------------------------------------------------------------------

If you prefer the classical formulation of the completion law,
well... here it is then.

-----------------------------------------------------------------------------*)

val Completion_thm_classical
	=
	store_thm
	("Completion_thm_classical",
	--`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies unless /\
	   isSubRelationOf implies ensures /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   satisfiesUNLESS_Disj unless /\
	   satisfiesPSP ensures unless /\
	   hasProperDomain (ensures:^UOP) /\  (* *)
	   (!p. inDomain ensures p ==> inDomain implies p) /\ (* *)
	   satisfiesCompLemma unless ensures (GLEADSTO ensures) /\

	   GLEADSTO ensures p (q OR b) /\ unless q b /\
	   GLEADSTO ensures r (s OR b) /\ unless s b 
	   ==>
           GLEADSTO ensures (p AND r) (q AND s OR  b)`--,

	   STRIP_TAC
	   THEN ONCE_REWRITE_TAC
		[prove(--`q AND s OR b = (q OR b) AND (s OR b) OR b`--,
			pred_eq_PROVE_TAC)
		]
	   THEN MATCH_MP_TAC Completion_preparation5
	   THEN RW_TAC std_ss []
	   THEN (CONV_TAC o RAND_CONV o REWR_CONV o prove)
			(--`b = b OR b`--, pred_eq_PROVE_TAC)
	   THEN MATCH_MP_TAC unless_simple_disjunction_thm
	   THEN RW_TAC std_ss []
	   THEN PROVE_TAC [unless_REFL_thm,
			   hasProperDomain_def]
	) ;


(*-----------------------------------------------------------------------------

Convergence. It is a generalization of self-stabilization (Pra95), and
an instance of COMPLETES.

Convergence is very nice to have because it is both disjunctive and
conjunctive.

-----------------------------------------------------------------------------*)

val CONVERGES_def = Define
	`CONVERGES leadsto unless p q
	 =
	 COMPLETES leadsto unless p q FF` ;


val Convergence_Conjunction_thm
	=
	store_thm
	("Convergence_Conjunction_thm",
	 --`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies unless /\
	   isSubRelationOf implies ensures /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   satisfiesUNLESS_Disj unless /\
	   satisfiesPSP ensures unless /\
	   hasProperDomain (ensures:^UOP) /\  (* *)
	   (!p. inDomain ensures p ==> inDomain implies p) /\ (* *)
	   satisfiesCompLemma unless ensures (GLEADSTO ensures) /\

	   CONVERGES (GLEADSTO ensures) unless p q /\
	   CONVERGES (GLEADSTO ensures) unless r s  
	   ==>
           CONVERGES (GLEADSTO ensures) unless 
		     (p AND r) 
		     (q AND s)`--,

	  RW_TAC std_ss [CONVERGES_def]
	  THEN (SUBST1_TAC o prove)
		(--`q AND s = q AND s OR (FF:^Pred)`--, pred_eq_PROVE_TAC)
	  THEN MATCH_MP_TAC Completion_thm
	  THEN RW_TAC std_ss [] ) ;

(*-----------------------------------------------------------------------------

A simpler (easier) to fulfill, but less general condition to obtain
the completion law.

-----------------------------------------------------------------------------*)

(*

val satisfies_immediatePSP_def = Define
	`satisfies_immediatePSP (ensures:^UOP) (unless:^UOP)
	 =
	(!p q a b. ensures p q /\ unless a b 
		   ==> 
		   ensures (p AND a AND NOT b) (q AND (a OR b)))`
	;

*)

val satisfiesSimplerCompLemma_def = Define
	`satisfiesSimplerCompLemma ensures 
	 =
	 !p q r s. 
		(ensures:^UOP) p q /\
	 	ensures r s 
		==>
		GLEADSTO ensures (p AND r) ((q AND (r OR s)) OR (p AND s))` ;
	;


val implies_domain_TAC 
	=
	FIRST (map MATCH_MP_TAC 
		(domain_closed_qOR_lemma :: domain_closedness_lemmas))
	THEN
	ASM_REWRITE_TAC [] 
	THEN REPEAT CONJ_TAC ;


val implies_satisfiesSimplerCompLemma
	=
	prove
	(--`includesIMP implies /\ 
	    isClosed_under_PredOPS (inDomain implies) /\ 
    	    isSubRelationOf implies ensures /\ 
	    hasProperDomain ensures /\ 
      	    (!p. inDomain ensures p ==> inDomain implies p) /\ 
	    isSubRelationOf ensures (unless:^UOP) /\
	    satisfiesUNLESS_Conj unless /\
	    satisfiesPSP ensures unless
	    ==>
	    satisfiesSimplerCompLemma ensures`--,

	  RW_TAC std_ss [isSubRelationOf_def,  satisfiesPSP_def,
			 satisfiesSimplerCompLemma_def]
	  THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_transitive_thm_morph1)
	  THEN EXISTS_TAC (--`(q AND (p AND r)) OR (q AND r OR s AND p OR q AND s)`--)
	  THEN CONJ_TAC
	  THENL [ALL_TAC,
		 MATCH_MP_TAC IMP_to_GLEADSTO_lift_thm
		 THEN RW_TAC std_ss [isSubRelationOf_def]
		 THENL [ALL_TAC, ALL_TAC, valid_PROVE_TAC]
		 THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
		 THEN RES_TAC
		 THEN RES_TAC
		 THEN REPEAT implies_domain_TAC ]
	 THEN MATCH_MP_TAC ensures_to_GLEADSTO_lift_thm
	 THEN (CONV_TAC o RATOR_CONV o RAND_CONV o REWR_CONV o prove)
		(--`p AND (r:^Pred) = p AND (p AND r)`--,
		    pred_eq_PROVE_TAC)
	 THEN FIRST_ASSUM MATCH_MP_TAC
	 THEN PROVE_TAC [ satisfiesUNLESS_Conj_def]
	) ;



val NORMALIZE_CONV =
	(REPEATC o CHANGED_CONV) 
		(DEPTH_CONV RIGHT_IMP_EXISTS_CONV
		 THENC 
		 (DEPTH_CONV RIGHT_AND_EXISTS_CONV)) ;


val simplerCompLemma_imp_CompLemma 
	= 
	prove
	(--`includesIMP implies /\
            isClosed_under_PredOPS (inDomain implies) /\
	    hasProperDomain unless /\
   	    isSubRelationOf implies ensures /\
	    hasProperDomain (ensures:^UOP) /\
	    (!p. inDomain unless p ==> inDomain implies p) /\
	    (!p. inDomain ensures p ==> inDomain implies p) /\
	    isSubRelationOf ensures (unless:^UOP) /\
	    satisfiesUNLESS_Conj unless /\
	    satisfiesPSP ensures unless 
	    ==>
 	    satisfiesCompLemma unless ensures (GLEADSTO ensures)`--,


	REWRITE_TAC [satisfiesCompLemma_def]
	THEN STRIP_TAC
	THEN SUBGOAL_THEN
		(--`satisfiesSimplerCompLemma (ensures:^UOP)`--)
		ASSUME_TAC
	THENL [PROVE_TAC [implies_satisfiesSimplerCompLemma],
	       ALL_TAC ]
	THEN REPEAT GEN_TAC
	THEN SUBGOAL_THEN
	 	(--`(ensures:^UOP) p q
		    ==>
		    isSubRelationOf 
			(GLEADSTO ensures) 
			(\ r s. GLEADSTO ensures r s /\
			        (unless s b ==>
    				 ?mid.
      				    GLEADSTO ensures (p AND r) (q AND mid OR q AND s OR b) /\
      				    GLEADSTO ensures mid s)
			)
		`--)
		(ASSUME_TAC o BETA_RULE o REWRITE_RULE [isSubRelationOf_def])
	THENL [ALL_TAC, PROVE_TAC []]
	THEN STRIP_TAC
	THEN REWRITE_TAC [GLEADSTO_equal_GLEADSTO2_thm]
	THEN MATCH_MP_TAC (GEN_ALL GLEADSTO2_induction1_thm)
	THEN REWRITE_TAC [SYM GLEADSTO_equal_GLEADSTO2_thm]
	THEN REPEAT CONJ_TAC
	THENL
	[ 
          (** (1) ensures lifting **)

	  RW_TAC std_ss [isSubRelationOf_def]
	  THENL [PROVE_TAC [ensures_to_GLEADSTO_lift_thm], ALL_TAC ]
	  THEN EXISTS_TAC (--`x:^Pred`--)
	  THEN RW_TAC std_ss [ensures_to_GLEADSTO_lift_thm]
	  THEN ONCE_REWRITE_TAC 
		[prove(--`q AND x OR q AND y OR b
			  =
			  (q AND x OR q AND y) OR (q AND y OR b)`--,
			pred_eq_PROVE_TAC) ]
	  THEN MATCH_MP_TAC (GEN (--`r:^Pred`--) GLEADSTO_Cancelation_thm)
	  THEN EXISTS_TAC (--`y AND (p:^Pred)`--)
	  THEN RW_TAC std_ss []
	  THENL
	    [ (* 1.1 *)
	      REWRITE_TAC [inDomain_def]
	      THEN MATCH_MP_TAC IMP_to_GLEADSTO_lift_thm
	      THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
	      THEN RES_TAC THEN RES_TAC
	      THEN RW_TAC std_ss []
	      THENL [ALL_TAC, ALL_TAC, valid_PROVE_TAC]
	      THEN PROVE_TAC domain_closedness_lemmas,
	      (* 1.2 *)
	      PROVE_TAC 
		[satisfiesSimplerCompLemma_def,
		 ensures_to_GLEADSTO_lift_thm,
		 prove(--`q AND x OR q AND y OR y AND p =
			  q AND (x OR y) OR p AND y`--,
			pred_eq_PROVE_TAC)
		],
	      (* 1.3 *)
	      SUBST1_TAC (prove(--`y AND p = p AND (y:^Pred)`--,
				pred_eq_PROVE_TAC))
	      THEN MATCH_MP_TAC PSP_thm_x 
	      THEN RW_TAC std_ss [ensures_to_GLEADSTO_lift_thm]
	    ],

	  (** (2) transitivity **)

	  RW_TAC std_ss [isTransitive2_def]
	  THENL [PROVE_TAC [ensures_to_GLEADSTO_lift_thm, 
			    GLEADSTO_transitive_thm_morph1],
		 ALL_TAC ]
	  THEN RES_TAC
	  THEN EXISTS_TAC (--`p' OR (q':^Pred) OR mid`--)
	  THEN CONJ_TAC
	  THENL
	  [ (** 2.1 **)
	    ONCE_REWRITE_TAC 
		[ prove(--`q AND (p' OR q' OR mid) OR q AND r OR b
			   =
			   q AND (p' OR q') OR
                           (q AND mid OR q AND r OR b)`--,
			pred_eq_PROVE_TAC)]
	    THEN MATCH_MP_TAC (GEN (--`r:^Pred`--) GLEADSTO_Cancelation_thm)
	    THEN EXISTS_TAC (--`p AND (q':^Pred)`--)
	    THEN RW_TAC std_ss []
	    THENL
	    	[ REWRITE_TAC [inDomain_def]
		  THEN MATCH_MP_TAC implies_to_GLEADSTO_lift_thm
		  THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
	      	  THEN PROVE_TAC (inDomain_def :: domain_closedness_lemmas),
		  ALL_TAC ]
	    THEN PROVE_TAC [ensures_to_GLEADSTO_lift_thm,
			    satisfiesSimplerCompLemma_def],

	    (** 2.2 **)

	    SUBST1_TAC (prove(--`(r:^Pred) = r OR r OR r`--,
			pred_eq_PROVE_TAC))
 	    THEN MATCH_MP_TAC GLEADSTO_SimpleDisjunction_thm
	    THEN RW_TAC std_ss []
	    THEN MATCH_MP_TAC GLEADSTO_SimpleDisjunction_thm
	    THEN RW_TAC std_ss []
	    THEN PROVE_TAC [ensures_to_GLEADSTO_lift_thm, 
			    GLEADSTO_transitive_thm_morph1]
	   ],

	  (** (3) disjunctivity **)

	  RW_TAC std_ss [isLeftDisj_def]
	  THENL [PROVE_TAC [GLEADSTO_LeftDisj_thm_morph1], ALL_TAC ]
	  THEN XRULE_ASSUM_TAC (CONV_RULE NORMALIZE_CONV)
	  THEN XRULE_ASSUM_TAC (CONV_RULE SKOLEM_CONV)
	  THEN UNDISCH_ALL_TAC THEN REPEAT STRIP_TAC
	  THEN EXISTS_TAC (--`??i:^Pred::W. (mid i):^Pred`--)
	  THEN CONJ_TAC
	  THENL [ALL_TAC, 
		 MATCH_MP_TAC (GEN_ALL GLEADSTO_LeftDisj_thm_morph2)
		 THEN RW_TAC std_ss[]
		 THEN PROVE_TAC [] ]
	  THEN ONCE_REWRITE_TAC 
		[prove(--`p AND (??p:^Pred::W. p) = (??i::W. (\i. p AND i) i)`--,
			pred_eq_PROVE_TAC)
		]
	  THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_LeftDisj_thm_morph2)
	  THEN EXISTS_TAC (--`p':^Pred`--) 
	  THEN ASM_REWRITE_TAC []
	  THEN REPEAT STRIP_TAC THEN BETA_TAC
	  THEN MATCH_MP_TAC (GEN_ALL GLEADSTO_transitive_thm_morph1)
	  THEN EXISTS_TAC (--`q AND mid (i:^Pred) OR q AND q' OR (b:^Pred)`--)
	  THEN CONJ_TAC
	  THENL [PROVE_TAC [], ALL_TAC ]
	  THEN MATCH_MP_TAC IMP_to_GLEADSTO_lift_thm
	  THEN ASM_REWRITE_TAC []
	  THEN REPEAT CONJ_TAC
	  THENL 
	   [ PROVE_TAC [hasProperDomain_def,
			GLEADSTO_hasProperDomain_lemma],
	     ALL_TAC, 
	     valid_PROVE_TAC
	   ]
	  THEN RES_TAC
	  THEN XRULE_ASSUM_TAC (REWRITE_RULE [hasProperDomain_def])
	  THEN RES_TAC THEN RES_TAC
	  THEN REPEAT implies_domain_TAC
	  THEN PROVE_TAC [hasProperDomain_def,
			  GLEADSTO_hasProperDomain_lemma]

	]
	) ;



(*-----------------------------------------------------------------------------

The completion laws, conditioned by the above easier condition

-----------------------------------------------------------------------------*)

val Completion_with_stricterCond_thm
	=
	store_thm
	("Completion_with_stricterCond_thm",
	--`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies unless /\
	   isSubRelationOf implies ensures /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   satisfiesUNLESS_Disj unless /\
	   hasProperDomain (ensures:^UOP) /\
	   (!p. inDomain ensures p ==> inDomain implies p) /\
	   isSubRelationOf ensures (unless:^UOP) /\
	   satisfiesPSP ensures unless /\
	   COMPLETES (GLEADSTO ensures) unless p q b /\
	   COMPLETES (GLEADSTO ensures) unless r s b 
	   ==>
           COMPLETES (GLEADSTO ensures) unless 
		     (p AND r) 
		     (q AND s OR  b)
		      b`--,

	   STRIP_TAC
	   THEN MATCH_MP_TAC Completion_thm
	   THEN RW_TAC std_ss []
	   THEN MATCH_MP_TAC simplerCompLemma_imp_CompLemma 
	   THEN RW_TAC std_ss []
	) ;	


val Completion_thm_classical_with_stricterCond
	=
	store_thm
	("Completion_thm_classical_with_stricterCond",
	--`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies unless /\
	   isSubRelationOf implies ensures /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   satisfiesUNLESS_Disj unless /\
	   hasProperDomain (ensures:^UOP) /\
	   (!p. inDomain ensures p ==> inDomain implies p) /\
	   isSubRelationOf ensures (unless:^UOP) /\
	   satisfiesPSP ensures unless /\

	   GLEADSTO ensures p (q OR b) /\ unless q b /\
	   GLEADSTO ensures r (s OR b) /\ unless s b 
	   ==>
           GLEADSTO ensures (p AND r) (q AND s OR  b)`--,

	   STRIP_TAC
	   THEN MATCH_MP_TAC Completion_thm_classical
	   THEN RW_TAC std_ss []
	   THEN MATCH_MP_TAC simplerCompLemma_imp_CompLemma 
	   THEN RW_TAC std_ss []
	) ;


val Convergence_Conjunction_thm_with_stricterCond
	=
	store_thm
	("Convergence_Conjunction_thm_with_stricterCond",
	 --`includesIMP implies /\
	   (!p. inDomain unless p ==> inDomain implies p) /\
	   isClosed_under_PredOPS (inDomain implies) /\
	   hasProperDomain unless /\
	   isSubRelationOf implies unless /\
	   isSubRelationOf implies ensures /\
	   satisfiesUNLESS_Subst implies (unless:^UOP) /\
	   satisfiesUNLESS_Conj unless /\
	   satisfiesUNLESS_Disj unless /\
	   hasProperDomain (ensures:^UOP) /\
	   (!p. inDomain ensures p ==> inDomain implies p) /\
	   isSubRelationOf ensures (unless:^UOP) /\
	   satisfiesPSP ensures unless /\

	   CONVERGES (GLEADSTO ensures) unless p q /\
	   CONVERGES (GLEADSTO ensures) unless r s  
	   ==>
           CONVERGES (GLEADSTO ensures) unless 
		     (p AND r) 
		     (q AND s)`--,
	   STRIP_TAC
	   THEN MATCH_MP_TAC Convergence_Conjunction_thm
	   THEN RW_TAC std_ss []
	   THEN MATCH_MP_TAC simplerCompLemma_imp_CompLemma 
	   THEN RW_TAC std_ss []
	) ;

(*-----------------------------------------------------------------------------

Conservative extension

-----------------------------------------------------------------------------*)

val isConservative_def = 
	Define 
	`isConservative (unityOp:'asm->^UOP) extend
	 =
	 !A p q. unityOp A p q ==> unityOp (extend A) p q` ;

val conservative_extension_thm
	=
	store_thm
	("conservative_extension_thm",
	--`isConservative (unityOp:'asm->^UOP) extend /\
	   unityOp A p q
	   ==>
	   unityOp (extend A) p q`--,
	PROVE_TAC [isConservative_def]) ;

val GLEADSTO_conservative_extension_thm
	=
	store_thm
	("GLEADSTO_conservative_extension_thm",
	--`isConservative (ensures_:'asm->^UOP) extend /\
	   GLEADSTO (ensures_ A) p q
	   ==>
	   GLEADSTO (ensures_ (extend A)) p q`--,

	REWRITE_TAC [isConservative_def]
	THEN STRIP_TAC
	THEN SUBGOAL_THEN
		(--`isSubRelationOf (GLEADSTO ((ensures_ :'asm->^UOP) A)) 
				    (GLEADSTO (ensures_ (extend A)))`--)
	     ASSUME_TAC
	THENL [ALL_TAC,
	       UNDISCH_ALL_TAC
	       THEN RW_TAC std_ss [isSubRelationOf_def]
	       THEN PROVE_TAC [] ]

	THEN PROVE_TAC [isSubRelationOf_def, GLEADSTO_inclusion_thm]
	) ;	
	

(*-----------------------------------------------------------------------------

Property on additional assumptions...

-----------------------------------------------------------------------------*)

val implicitlyImplies_def = 
	Define 
	`implicitlyImplies (unityOp:'asm->^UOP) property
	 =
	 !A p q. unityOp A p q ==> property A` ;

val implicitlyImplies_thm 
	=
	store_thm
	("implicitlyImplies_thm",
	--`implicitlyImplies (unityOp:'asm->^UOP) property /\
	   unityOp A p q
	   ==>
	   property A`--,

	PROVE_TAC [implicitlyImplies_def]) ;


val GLEADSTO_implicitlyImplies_thm
	=
	store_thm
	("GLEADSTO_implicitlyImplies_thm",
	--`implicitlyImplies (ensures_:'asm->^UOP) property  /\
	   GLEADSTO (ensures_ A) p q
	   ==>
	   property A`--,


	REWRITE_TAC [implicitlyImplies_def]
	THEN STRIP_TAC
	THEN SUBGOAL_THEN (--`isSubRelationOf (GLEADSTO ((ensures_ : 'asm->^UOP) A))
					 (\p q. property A)`--)
		ASSUME_TAC
	THENL [ALL_TAC, 
	       UNDISCH_ALL_TAC 
	       THEN RW_TAC std_ss [isSubRelationOf_def]
	       THEN PROVE_TAC []]
	THEN MATCH_MP_TAC GLEADSTO_induction1_thm
	THEN REPEAT CONJ_TAC
	THEN RW_TAC std_ss [isSubRelationOf_def, 
			    isTransitive_def, 
			    isLeftDisj_def]
	THEN PROVE_TAC []
	) ;


val _ = export_theory () ;

